// BenLincoln.ActiveMemory
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.ActiveMemory.

// BenLincoln.ActiveMemory is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.ActiveMemory is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.ActiveMemory (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Runtime.InteropServices;

namespace BenLincoln.ActiveMemory
{
    public class MemoryAccessor
    {
        #region Enum

        [Flags]
        public enum ProcessAccessRight
        {
            PROCESS_TERMINATE = (0x0001),
            PROCESS_CREATE_THREAD = (0x0002),
            PROCESS_SET_SESSIONID = (0x0004),
            PROCESS_VM_OPERATION = (0x0008),
            PROCESS_VM_READ = (0x0010),
            PROCESS_VM_WRITE = (0x0020),
            PROCESS_DUP_HANDLE = (0x0040),
            PROCESS_CREATE_PROCESS = (0x0080),
            PROCESS_SET_QUOTA = (0x0100),
            PROCESS_SET_INFORMATION = (0x0200),
            PROCESS_QUERY_INFORMATION = (0x0400),
            PROCESS_QUERY_LIMITED_INFORMATION = (0x1000),
            PROCESS_ALL_ACCESS = (0x1F0FFF)
        }

        #endregion

        public static uint GetProcessIDByName(string processName)
        {
            Process.EnterDebugMode();
            Process[] processes = Process.GetProcessesByName(processName);
            Process.LeaveDebugMode();

            if (processes.GetUpperBound(0) < 0)
            {
                throw new ProcessNotFoundException("Unable to find the process named '" + processName + "'.");
            }

            if (processes.GetUpperBound(0) > 0)
            {
                throw new ProcessNameIsNotUniqueException("There is more than one process named '" + processName + 
                    "' running concurrently. Please use the process ID instead of the name.");
            }

            return (uint)processes[0].Id;
        }

        public static byte[] ReadMemory(uint processID, uint startAddress, uint numBytes)
        {
            ProcessAccessRight readRight = ProcessAccessRight.PROCESS_VM_READ;

            byte[] readBuffer = new byte[numBytes];
            IntPtr bytesRead;

            try
            {
                IntPtr process = OpenProcess((uint)readRight, 1, processID);

                ReadProcessMemory(process, (IntPtr)startAddress, readBuffer, numBytes, out bytesRead);

                CloseHandle(process);
            }
            catch (Exception ex)
            {
                throw new MemoryReadException("An error occurred reading memory from the process.", ex);
            }

            return readBuffer;
        }

        public static void WriteMemory(uint processID, uint startAddress, byte[] writeBuffer)
        {
            ProcessAccessRight writeRight = ProcessAccessRight.PROCESS_VM_WRITE | ProcessAccessRight.PROCESS_VM_OPERATION;

            IntPtr bytesWritten;

            try
            {
                IntPtr process = OpenProcess((uint)writeRight, 1, processID);

                WriteProcessMemory(process, (IntPtr)startAddress, writeBuffer, (uint)writeBuffer.Length, out bytesWritten);

                CloseHandle(process);
            }
            catch (Exception ex)
            {
                throw new MemoryWriteException("An error occurred writing to the process memory.", ex);
            }
        }


        #region Imported From kernel32.dll

        /* This section is based on Sojaner's memory scanner application.
         * He in turn credited "Arik Poznanski for P/Invokes and methods needed to read and write the Memory
         * For more information refer to "Minesweeper, Behind the scenes" article by Arik Poznanski at Codeproject.com".
         */

        //		HANDLE OpenProcess(
        //			DWORD dwDesiredAccess,  // enumerated as ProcessAccessRight, above
        //			BOOL bInheritHandle,    // processes created by this process inherit the handle
        //			DWORD dwProcessId       // process ID
        //			);
        [DllImport("kernel32.dll")]
        protected static extern IntPtr OpenProcess(UInt32 dwDesiredAccess, Int32 bInheritHandle, UInt32 dwProcessId);

        //		BOOL CloseHandle(
        //			HANDLE hObject   // object handle
        //			);
        [DllImport("kernel32.dll")]
        protected static extern Int32 CloseHandle(IntPtr hObject);

        //		BOOL ReadProcessMemory(
        //			HANDLE hProcess,              // process handle
        //			LPCVOID lpBaseAddress,        // pointer to base memory address
        //			LPVOID lpBuffer,              // output data buffer
        //			SIZE_T nSize,                 // number of bytes to read
        //			SIZE_T * lpNumberOfBytesRead  // number of bytes read
        //			);
        [DllImport("kernel32.dll")]
        protected static extern Int32 ReadProcessMemory(IntPtr hProcess, IntPtr lpBaseAddress, [In, Out] byte[] buffer, UInt32 size, out IntPtr lpNumberOfBytesRead);

        //		BOOL WriteProcessMemory(
        //			HANDLE hProcess,                // process handle
        //			LPVOID lpBaseAddress,           // pointer to base memory address
        //			LPCVOID lpBuffer,               // input data buffer
        //			SIZE_T nSize,                   // number of bytes to write
        //			SIZE_T * lpNumberOfBytesWritten // number of bytes written
        //			);
        [DllImport("kernel32.dll")]
        protected static extern Int32 WriteProcessMemory(IntPtr hProcess, IntPtr lpBaseAddress, [In, Out] byte[] buffer, UInt32 size, out IntPtr lpNumberOfBytesWritten);

        #endregion

    }
}
