// BenLincoln.ActiveMemory
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.ActiveMemory.

// BenLincoln.ActiveMemory is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.ActiveMemory is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.ActiveMemory (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.ActiveMemory
{
    public class MemoryFragment
    {
        protected uint _Address;
        protected byte[] _Values;

        #region Properties

        public uint Address
        {
            get
            {
                return _Address;
            }
        }

        public byte[] Values
        {
            get
            {
                return _Values;
            }
        }

        #endregion

        public MemoryFragment(uint address, byte[] values)
        {
            _Address = address;
            _Values = values;
        }

        public MemoryFragment(uint address, string values)
        {
            _Address = address;

            if (values.Length == 0)
            {
                throw new FormatException("Value string cannot be empty");
            }

            if ((values.Length % 2) == 1)
            {
                throw new FormatException("Value string must have an even number of characters");
            }

            string hexChars = "[0-9a-fA-F]";

            Regex rex = new Regex(hexChars);

            if (!rex.IsMatch(values))
            {
                throw new FormatException("'" + values + "' is not a valid set of bytes represented in hexadecimal.");
            }

            // now that the input has been validated, split it into groups of two characters and parse those into bytes

            byte[] hexVals = new byte[values.Length / 2];

            string workingVals = values;

            int byteNum = 0;

            while (workingVals.Length > 0)
            {
                string currentVal = workingVals.Substring(0, 2);
                workingVals = workingVals.Substring(2);
                hexVals[byteNum] = byte.Parse(currentVal, System.Globalization.NumberStyles.AllowHexSpecifier);
                byteNum++;
            }

            _Values = hexVals;
        }

        public bool Equals(MemoryFragment frag)
        {
            if (frag.Address != _Address)
            {
                return false;
            }
            return Equals(frag.Values);
        }

        public bool Equals(byte[] fragValues)
        {
            if (fragValues == null)
            {
                return false;
            }
            if (fragValues.GetUpperBound(0) != _Values.GetUpperBound(0))
            {
                return false;
            }
            bool isEqual = true;
            for (int i = 0; i < (_Values.GetUpperBound(0) + 1); i++)
            {
                if (fragValues[i] != _Values[i])
                {
                    isEqual = false;
                }
            }
            return isEqual;
        }
    }
}
