// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class SR1PCPMFClip : BLCDDA.SoundClip
    {
        protected const long _HeaderLength = 8;

        public SR1PCPMFClip(string name, BLCDDA.SoundClipContainerFile parentFile, long offset, long length, uint sampleRate)
            : base(name, parentFile, offset, length - _HeaderLength)
        {
            _NumChannels = 1;
            _SampleRate = sampleRate;
            _BitDepth = 16;
        }

        public override void LoadFileData()
        {
            uint numSamplesPerChannel = (uint)((_Length / 2) / _NumChannels);    // /2 is for 16 bits
            FileStream inStream = new FileStream(_ParentSoundClipContainerFile.Path, FileMode.Open, FileAccess.Read);
            inStream.Seek(_Offset, SeekOrigin.Begin);
            // check for loop points
            BinaryReader iReader = new BinaryReader(inStream);
            long loopStart = iReader.ReadUInt32();
            long loopEnd = iReader.ReadUInt32();

            if ((loopStart > 0) || (loopEnd > 0))
            {
                BLDA.LoopPoint loop = new BLDA.LoopPoint(loopStart, loopEnd);
                _LoopPoints = new BLDA.LoopPoint[1] { loop };
            }

            // now get the actual audio data
            _WAVFile = new BLDA.WAVFile((ushort)_NumChannels, (uint)_SampleRate, (ushort)_BitDepth, numSamplesPerChannel);
            // have the WAV file load the rest of the data in a thread
            Thread loadThread = new Thread(_WAVFile.ReadDataThreaded);
            loadThread.Start(inStream);
            // wait for the loading to finish, and keep the load value in this class updated
            do
            {
                _IOPercent = _WAVFile.Progress();
                Thread.Sleep(5);
            }
            while (loadThread.IsAlive);

            inStream.Close();
        }

        protected override Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "Soul Reaver (PC) PCM audio");
            return info;
        }
    }
}
