// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class SR1PCPMFFile : BLCDDA.SoundClipContainerFile
    {
        public SR1PCPMFFile(string path)
            : base(path)
            {
            }

        public override void LoadFile()
        {
            FileStream scanStream = new FileStream(_Path, FileMode.Open, FileAccess.Read);
            long currentClipStart = 0;
            long currentPosition = 0;
            long streamLength = scanStream.Length;
            _IOPercent = 0;

            string[] pathSplit = _Path.Split('\\');
            string fileName = pathSplit[pathSplit.GetUpperBound(0)];
            fileName = fileName.Substring(0, fileName.LastIndexOf('.'));

            BinaryReader iReader = new BinaryReader(scanStream);

            // verify using the header that this is in fact a PMF file
            uint header = iReader.ReadUInt32();
            if (header != 0x61504D46)
            {
                throw new BLS.InvalidFileFormatException("The file does not have the 'FMPa' header");
            }
            // don't know what this is yet
            uint unknown1 = iReader.ReadUInt32();
            // get number of clips inside
            int clipCount = (int)iReader.ReadUInt32();
            // don't know what this is yet
            uint unknown2 = iReader.ReadUInt32();

            currentClipStart = 16;
            currentPosition = 16;

            _SoundClips = new BLCDDA.SoundClip[clipCount];

            for (int clipNum = 0; clipNum < clipCount; clipNum++)
            {
                // skip the first 20 bytes, don't know what they're for yet
                currentPosition += 20;
                scanStream.Seek(currentPosition, SeekOrigin.Begin);

                // read the length of the data
                long currentClipLength = iReader.ReadUInt32() - 4;
                // read the sample rate
                uint sampleRate = iReader.ReadUInt32();
                currentClipStart = scanStream.Position;
                _SoundClips[clipNum] = new BLCDDA.SR1PCPMFClip(fileName + string.Format("-{0:0000}", clipNum), this, currentClipStart, currentClipLength, sampleRate);
                currentPosition = currentClipStart + currentClipLength;
            }

            iReader.Close();
            scanStream.Close();

            _IOPercent = 1;

        }

        protected override Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "Soul Reaver (PC) Sound Collection");
            return info;
        }
    }
}
