// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class WAVClip : BLCDDA.SoundClip
    {

        public WAVClip(string name, BLCDDA.SoundClipContainerFile parentFile, long offset)
            : base(name, parentFile, offset)
        {
        }

        public override void LoadFileData()
        {
            FileStream inStream = new FileStream(_ParentSoundClipContainerFile.Path, FileMode.Open, FileAccess.Read);
            inStream.Seek(_Offset, SeekOrigin.Begin);
            _WAVFile = BLDA.WAVFile.FromStreamHeaderOnly(inStream);
            // have the WAV file load the rest of the data in a thread
            Thread loadThread = new Thread(_WAVFile.ReadDataThreaded);
            loadThread.Start(inStream);
            // wait for the loading to finish, and keep the load value in this class updated
            do
            {
                _IOPercent = _WAVFile.Progress();
                Thread.Sleep(5);
            }
            while (loadThread.IsAlive);


            inStream.Close();

            // set class variables
            _Length = _WAVFile.HeaderLength + _WAVFile.SubChunk2Size;
            _SampleRate = _WAVFile.SampleRate;
            _BitDepth = _WAVFile.BitsPerSample;
            _NumChannels = _WAVFile.NumChannels;
        }

        protected override Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "RIFF WAV File");
            info.Add("Chunk ID", _WAVFile.ChunkID.ToString());
            info.Add("Chunk Size", _WAVFile.ChunkSize.ToString());
            info.Add("Format", _WAVFile.Format.ToString());
            info.Add("Sub-Chunk 1 ID", _WAVFile.SubChunk1ID.ToString());
            info.Add("Sub-Chunk 1 Size", _WAVFile.SubChunk1Size.ToString() + " bytes");
            info.Add("Audio Format", _WAVFile.AudioFormat.ToString());
            info.Add("Byte Rate", _WAVFile.ByteRate.ToString());
            info.Add("Block Align", _WAVFile.BlockAlign.ToString());
            info.Add("Sub-Chunk 2 ID", _WAVFile.SubChunk2ID.ToString());
            info.Add("Sub-Chunk 2 Size", _WAVFile.SubChunk2Size.ToString() + " bytes");
            return info;
        }
    }
}
