// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class ERGVariableBool : ERGVariable
    {
        protected bool _Value;
        protected bool _IsInvertedValue;

        #region Properties

        public bool Value
        {
            get
            {
                return _Value;
            }
        }

        public bool IsInvertedValue
        {
            get
            {
                return _IsInvertedValue;
            }
        }

        #endregion

        public ERGVariableBool(string category, string name, Control boundControl, bool initialValue, bool isInvertedValue)
            : base(category, name, boundControl)
        {
            if (boundControl != null)
            {
                CheckBox bound = GetBoundControlAsCheckBox();
                bound.CheckedChanged += new EventHandler(BoundControlHandler);
            }
            _Value = initialValue;
            _IsInvertedValue = isInvertedValue;
        }

        public override void BindControls(Control[] boundControls)
        {
            _BoundControls = boundControls;
            if (_BoundControls[0] != null)
            {
                CheckBox bound = GetBoundControlAsCheckBox();
                bound.CheckedChanged += new EventHandler(BoundControlHandler);
            }
            UpdateBoundControls();
        }

        public void BoundControlHandler(object sender, EventArgs e)
        {
            GetValueFromBoundControls();
        }

        protected CheckBox GetBoundControlAsCheckBox()
        {
            CheckBox bound;
            try
            {
                bound = (CheckBox)_BoundControls[0];
            }
            catch (InvalidCastException cEx)
            {
                throw new InvalidCastException("Boolean variables can only be bound to a control of type CheckBox.");
            }
            return bound;
        }

        public override string GetValue()
        {
            if (_IsInvertedValue)
            {
                if (_Value == true)
                {
                    return "0";
                }
                else
                {
                    return "1";
                }
            }
            else
            {
                if (_Value == true)
                {
                    return "1";
                }
                else
                {
                    return "0";
                }
            }
        }

        public static bool StringToBool(string value)
        {
            switch (value)
            {
                case "0":
                    return false;
                    break;
                case "1":
                    return true;
                    break;
                default:
                    throw new Exception("The value '" + value + "' is invalid for conversion to a game.erg boolean.");
                    break;
            }
        }

        public override void SetValue(string newValue, bool invertValue)
        {
            switch (newValue)
            {
                case "0":
                    _Value = false;
                    break;
                case "1":
                    _Value = true;
                    break;
                default:
                    throw new Exception("The value '" + newValue + "' is invalid for conversion to a game.erg boolean.");
                    break;
            }
            if (invertValue)
            {
                _Value = !_Value;
            }
            // yay for developers using the Bill and Ted model of booleans
            if (_IsInvertedValue)
            {
                _Value = !_Value;
            }
            UpdateBoundControls();
        }

        public override void UpdateBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                CheckBox bound = GetBoundControlAsCheckBox();
                if (_Value == true)
                {
                    bound.Checked = true;
                }
                else
                {
                    bound.Checked = false;
                }
            }
        }

        public override void GetValueFromBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                CheckBox bound = GetBoundControlAsCheckBox();
                if (bound.Checked == true)
                {
                    _Value = true;
                }
                else
                {
                    _Value = false;
                }
            }
        }
    }
}
