// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class ERGVariableInt : ERGVariable
    {
        protected int _Value;

        #region Properties

        public int Value
        {
            get
            {
                return _Value;
            }
        }

        #endregion

        public ERGVariableInt(string category, string name, Control boundControl, int initialValue)
            : base(category, name, boundControl)
        {
            if (boundControl != null)
            {
                TextBox bound = GetBoundControlAsTextBox();
                bound.TextChanged += new EventHandler(BoundControlHandler);
            }
            _Value = initialValue;
        }

        public override void BindControls(Control[] boundControls)
        {
            _BoundControls = boundControls;
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox();
                bound.TextChanged += new EventHandler(BoundControlHandler);
            }
            UpdateBoundControls();
        }

        public void BoundControlHandler(object sender, EventArgs e)
        {
            TextBox bound = GetBoundControlAsTextBox();
            if (bound.Text != "")
            {
                try
                {
                    GetValueFromBoundControls();
                }
                catch (FormatException ex)
                {
                    // do nothing, user is probably clearing the textbox to type something
                }
                catch (InvalidCastException ex)
                {
                    // do nothing, user is probably clearing the textbox to type something
                }
            }
        }

        public TextBox GetBoundControlAsTextBox()
        {
            TextBox bound;
            try
            {
                bound = (TextBox)_BoundControls[0];
            }
            catch (InvalidCastException cEx)
            {
                throw new InvalidCastException("Integer variables can only be bound to a control of type TextBox.");
            }
            return bound;
        }

        public override string GetValue()
        {
            return _Value.ToString();
        }

        public static int StringToInt(string value)
        {
            try
            {
                return int.Parse(value);
            }
            catch (Exception ex)
            {
                throw new Exception("The value '" + value + "' is invalid for conversion to a game.erg integer.");
            }
        }

        public override void SetValue(string newValue, bool invertValue)
        {
            try
            {
                _Value = int.Parse(newValue);
            }
            catch (Exception ex)
            {
                throw new Exception("The value '" + newValue + "' is invalid for conversion to a game.erg integer.");
            }
            UpdateBoundControls();
        }

        public override void UpdateBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox(); 
                bound.Text = GetValue();
            }
        }

        public override void GetValueFromBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox(); 
                SetValue(bound.Text);
            }
        }
    }
}
