using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class ImageData
    {
        protected byte[][] mChannelAlpha;
        protected byte[][] mChannelRed;
        protected byte[][] mChannelGreen;
        protected byte[][] mChannelBlue;

        protected Color[] mPalette;

        protected int mImageWidth;
        protected int mImageHeight;

        //protected ImageFile mImageFile;
        protected RAWImage _RAWImage;

        #region Properties

        public byte[][] ChannelAlpha
        {
            get
            {
                return mChannelAlpha;
            }
            set
            {
                mChannelAlpha = value;
            }
        }

        public byte[][] ChannelRed
        {
            get
            {
                return mChannelRed;
            }
            set
            {
                mChannelRed = value;
            }
        }

        public byte[][] ChannelGreen
        {
            get
            {
                return mChannelGreen;
            }
            set
            {
                mChannelGreen = value;
            }
        }


        public byte[][] ChannelBlue
        {
            get
            {
                return mChannelBlue;
            }
            set
            {
                mChannelBlue = value;
            }
        }

        public int ImageWidth
        {
            get
            {
                return mImageWidth;
            }
            set
            {
                mImageWidth = value;
            }
        }

        public int ImageHeight
        {
            get
            {
                return mImageHeight;
            }
            set
            {
                mImageHeight = value;
            }
        }

        #endregion

        //public ImageData(ImageFile inFile)
        //{
        //    mImageFile = inFile;
        //    ImageWidth = inFile.ImageWidth;
        //    ImageHeight = inFile.ImageHeight;
        //    InitializeChannels();
        //}

        //public ImageData(ImageFile inFile, int imgWidth, int imgHeight)
        //{
        //    mImageFile = inFile;
        //    ImageWidth = imgWidth;
        //    ImageHeight = imgHeight;
        //    InitializeChannels();
        //}

        public ImageData(RAWImage inFile)
        {
            _RAWImage = inFile;
            ImageWidth = inFile.Width;
            ImageHeight = inFile.Height;
            InitializeChannels();
        }

        public ImageData(RAWImage inFile, int imgWidth, int imgHeight)
        {
            _RAWImage = inFile;
            ImageWidth = imgWidth;
            ImageHeight = imgHeight;
            InitializeChannels();
        }

        protected void InitializeChannels()
        {
            mChannelAlpha = new byte[ImageWidth][];
            for (int i = 0; i < ImageWidth; i++)
            {
                mChannelAlpha[i] = new byte[ImageHeight];
            }
            mChannelRed = new byte[ImageWidth][];
            for (int i = 0; i < ImageWidth; i++)
            {
                mChannelRed[i] = new byte[ImageHeight];
            }
            mChannelGreen = new byte[ImageWidth][];
            for (int i = 0; i < ImageWidth; i++)
            {
                mChannelGreen[i] = new byte[ImageHeight];
            }
            mChannelBlue = new byte[ImageWidth][];
            for (int i = 0; i < ImageWidth; i++)
            {
                mChannelBlue[i] = new byte[ImageHeight];
            }
        }

        public void GetChannelsFromRawBytes(byte[] inData, bool ignoreAlpha)
        {
            GetChannelsFromRawBytes(inData, _RAWImage.BPPImage,
                _RAWImage.BPPAlpha, _RAWImage.BPPRed, _RAWImage.BPPGreen, _RAWImage.BPPBlue,
                _RAWImage.StartBitAlpha, _RAWImage.StartBitRed, _RAWImage.StartBitGreen, _RAWImage.StartBitBlue, ignoreAlpha);
        }

        public void GetChannelsFromRawBytes(byte[] inData, int bppImage,
            int bppAlpha, int bppRed, int bppGreen, int bppBlue,
            int startBitAlpha, int startBitRed, int startBitGreen, int startBitBlue, bool ignoreAlpha)
        {
            ulong pixelBuffer, bufferTemp;
            int pixelOmeter, globalByteOmeter, bytesPerPixel, x, y;
            byte tempAlpha, tempRed, tempGreen, tempBlue;
            ushort bufferLength;
            ushort factorAlpha, factorRed, factorGreen, factorBlue;

            //get app settings
            //ProgramSettings currentSettings = new ProgramSettings();
            //bool ignoreAlpha = currentSettings.IgnoreAlpha;

            //are we ignoring the alpha data?
            if ((ignoreAlpha != null) && (ignoreAlpha == true))
            {
                bppAlpha = 0;
                startBitAlpha = 0;
            }

            bytesPerPixel = bppImage / 8;

            factorAlpha = (ushort)(8 - bppAlpha);
            factorRed = (ushort)(8 - bppRed);
            factorGreen = (ushort)(8 - bppGreen);
            factorBlue = (ushort)(8 - bppBlue);

            globalByteOmeter = 0;
            x = 0;
            y = 0;

            pixelBuffer = 0;
            bufferLength = 0;
            for (pixelOmeter = 0; pixelOmeter < mImageWidth * mImageHeight; pixelOmeter++)
            {
                bufferTemp = 0;

                //if the buffer does not have enough bits in it to represent the current
                //pixel, read in another byte
                if (bufferLength < bppImage)
                {
                    do
                    {
                        bufferTemp = inData[globalByteOmeter];
                        pixelBuffer = pixelBuffer << 8;
                        pixelBuffer += bufferTemp;
                        bufferLength += 8;
                        globalByteOmeter++;
                    }
                    while (bufferLength < bppImage);
                }

                //circular shift for 16-bit images
                if (bppImage == 16)
                {
                    ulong tBuffer1 = pixelBuffer;
                    ulong tBuffer2 = pixelBuffer;
                    tBuffer1 <<= 8;
                    tBuffer1 &= 0xFF00;
                    tBuffer2 >>= 8;
                    tBuffer2 &= 0xFF;
                    pixelBuffer = tBuffer1 | tBuffer2;
                }

                //read the colour data from the pixel buffer into the channels
                //converting it to 16-bit-per-channel colour in the process

                //Alpha
                if ((bppAlpha == 0) || ignoreAlpha)
                {
                    tempAlpha = 255;
                }
                else
                {
                    bufferTemp = pixelBuffer;
                    //if GIYStart is greater than 0, trim off the excess
                    bufferTemp = bufferTemp << (64 - bufferLength + startBitAlpha);
                    //trim off the right side of the buffer
                    bufferTemp = bufferTemp >> (64 - bppAlpha);
                    tempAlpha = (byte)(bufferTemp << factorAlpha);
                }

                //premultFactor = (float)256 / (float)tempAlpha;

                //Red
                if (bppRed == 0)
                {
                    tempRed = 0;
                }
                else
                {
                    bufferTemp = pixelBuffer;
                    //if GIYStart is greater than 0, trim off the excess
                    bufferTemp = bufferTemp << (64 - bufferLength + startBitRed);
                    //trim off the right side of the buffer
                    bufferTemp = bufferTemp >> (64 - bppRed);
                    tempRed = (byte)(bufferTemp << factorRed);
                }

                //Green
                if (bppGreen == 0)
                {
                    tempGreen = 0;
                }
                else
                {
                    bufferTemp = pixelBuffer;
                    //if GIYStart is greater than 0, trim off the excess
                    bufferTemp = bufferTemp << (64 - bufferLength + startBitGreen);
                    //trim off the right side of the buffer
                    bufferTemp = bufferTemp >> (64 - bppGreen);
                    tempGreen = (byte)(bufferTemp << factorGreen);
                }

                //Blue
                if (bppBlue == 0)
                {
                    tempBlue = 0;
                }
                else
                {
                    bufferTemp = pixelBuffer;
                    //if GIYStart is greater than 0, trim off the excess
                    bufferTemp = bufferTemp << (64 - bufferLength + startBitBlue);
                    //trim off the right side of the buffer
                    bufferTemp = bufferTemp >> (64 - bppBlue);
                    tempBlue = (byte)(bufferTemp << factorBlue);
                }

                //now remove this pixel's data from the main buffer
                bufferLength -= (ushort)bppImage;
                pixelBuffer = pixelBuffer << (64 - bufferLength);
                pixelBuffer = pixelBuffer >> (64 - bufferLength);

                mChannelAlpha[x][y] = (byte)tempAlpha;

                //if (alphaPremultiplied)
                //{
                //    tempRed = (byte)((float)tempRed * premultFactor);
                //    tempGreen = (byte)((float)tempGreen * premultFactor);
                //    tempBlue = (byte)((float)tempBlue * premultFactor);
                //}

                mChannelRed[x][y] = (byte)tempRed;
                mChannelGreen[x][y] = (byte)tempGreen;
                mChannelBlue[x][y] = (byte)tempBlue;

                // increment position in grid
                x++;
                if (x == mImageWidth)
                {
                    y++;
                    x = 0;
                }
            }
        }

        public void Get4BitChannelsFromRawBytes(byte[] inData)
        {
            ulong pixelBuffer, bufferTemp;
            int pixelOmeter, globalByteOmeter, bytesPerPixel, x, y;
            byte tempLuma;
            ushort bufferLength;
            ushort factorLuma;

            //get app settings
            //ProgramSettings currentSettings = new ProgramSettings();

            factorLuma = 4;

            globalByteOmeter = 0;
            x = 0;
            y = 0;

            pixelBuffer = 0;
            bufferLength = 0;
            for (pixelOmeter = 0; pixelOmeter < mImageWidth * mImageHeight; pixelOmeter++)
            {
                bufferTemp = 0;

                //if the buffer does not have enough bits in it to represent the current
                //pixel, read in another byte
                if (bufferLength < 4)
                {
                    bufferTemp = inData[globalByteOmeter];
                    pixelBuffer = pixelBuffer << 8;
                    pixelBuffer += bufferTemp;
                    bufferLength += 8;
                    globalByteOmeter++;
                    //endian change
                    ulong bufferBuffer = pixelBuffer;
                    pixelBuffer = pixelBuffer << 4;
                    bufferBuffer = bufferBuffer >> 4;
                    pixelBuffer &= 0xF0;
                    bufferBuffer &= 0x0F;
                    pixelBuffer |= bufferBuffer;
                }

                //read the colour data from the pixel buffer into the channels
                //converting it to 16-bit-per-channel colour in the process

                //Luma
                bufferTemp = pixelBuffer;
                //if GIYStart is greater than 0, trim off the excess
                bufferTemp = bufferTemp << (60);
                //trim off the right side of the buffer
                bufferTemp = bufferTemp >> (60);
                tempLuma = (byte)(bufferTemp << 4);

                //now remove this pixel's data from the main buffer
                bufferLength -= 4;
                pixelBuffer = pixelBuffer << (64 - bufferLength);
                pixelBuffer = pixelBuffer >> (64 - bufferLength);

                mChannelAlpha[x][y] = (byte)255;

                //if (alphaPremultiplied)
                //{
                //    tempRed = (byte)((float)tempRed * premultFactor);
                //    tempGreen = (byte)((float)tempGreen * premultFactor);
                //    tempBlue = (byte)((float)tempBlue * premultFactor);
                //}

                mChannelRed[x][y] = (byte)tempLuma;
                mChannelGreen[x][y] = (byte)tempLuma;
                mChannelBlue[x][y] = (byte)tempLuma;

                // increment position in grid
                x++;
                if (x == mImageWidth)
                {
                    y++;
                    x = 0;
                }
            }
        }

        public void GetPaletteFromBitmap(Bitmap inPalette)
        {
            mPalette = new Color[256];

            for (int i = 0; i < 256; i++)
            {
                mPalette[i] = inPalette.GetPixel(i, 0);
            }
        }

        public Bitmap CreateImage()
        {
            Bitmap createdImage = new Bitmap(mImageWidth, mImageHeight);

            for (int y = 0; y < mImageHeight; y++)
            {
                for (int x = 0; x < mImageWidth; x++)
                {
                    Color currentColour = Color.FromArgb
                        (
                        mChannelAlpha[x][y],
                        mChannelRed[x][y],
                        mChannelGreen[x][y],
                        mChannelBlue[x][y]
                        );
                    createdImage.SetPixel(x, y, currentColour);
                }
            }

            return createdImage;
        }

    }
}
