// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public struct SegmentedRAWParameters
    {
        public int SubImageWidth1;
        public int SubImageWidth2;
        public int SubImageWidth3;
        public int SubImageWidth4;
        public int SubImageWidth5;
        public int SubImageWidth6;
    }

    public abstract class SegmentedRAWImage : RAWImage
    {
        protected const int _SegmentationDataHeaderOffset = 84;

        protected SegmentedRAWParameters _SegmentedRAWParameters;

        protected int _SubImageHorizontalCount;
        protected int _SubImageVerticalCount;

        #region Properties

        public int SubImageHorizontalCount
        {
            get
            {
                return _SubImageHorizontalCount;
            }
        }

        public int SubImageVerticalCount
        {
            get
            {
                return _SubImageVerticalCount;
            }
        }

        public int SubImageCount
        {
            get
            {
                return (_SubImageHorizontalCount * _SubImageVerticalCount);
            }
        }

        #endregion

        public override string GetInfo()
        {
            StringBuilder builder = new StringBuilder(base.GetInfo());

            builder.Append("Number of Segments: ");
            builder.Append(SubImageCount);
            builder.Append(" (");
            builder.Append(SubImageHorizontalCount.ToString());
            builder.Append("x");
            builder.Append(SubImageVerticalCount.ToString());
            builder.Append(")");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        public static SegmentedRAWParameters GetSegmentedParameters(string path)
        {
            SegmentedRAWParameters segParms = new SegmentedRAWParameters();

            FileStream iStream = new FileStream(path, FileMode.Open, FileAccess.Read);
            BinaryReader iReader = new BinaryReader(iStream);

            iStream.Seek(_SegmentationDataHeaderOffset, SeekOrigin.Begin);

            segParms.SubImageWidth1 = (int)iReader.ReadUInt32();
            segParms.SubImageWidth2 = (int)iReader.ReadUInt32();
            segParms.SubImageWidth3 = (int)iReader.ReadUInt32();
            segParms.SubImageWidth4 = (int)iReader.ReadUInt32();
            segParms.SubImageWidth5 = (int)iReader.ReadUInt32();
            segParms.SubImageWidth6 = (int)iReader.ReadUInt32();

            iReader.Close();
            iStream.Close();

            return segParms;
        }

    }
}
