// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using BLI = BenLincoln.Imaging;
using BLPS2 = BenLincoln.Playstation2;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class TRLPSPIndexedRAWImage : RAWImage
    {
        protected const long _PaletteOffset = 128;
        protected const int _PaletteLength = 1024;
        protected const long _ImageOffset = 1152;

        public TRLPSPIndexedRAWImage(string path, RAWParameters parms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _TypeDescription = "Tomb Raider: Legend (PSP) RAW Bitmap";
            _FormatDescription = "Uncompressed Twiddled Indexed Image";

            _DataOffset = 128;

            //these values are hardcoded because the header information is incorrect in this filetype
            _RAWParameters.BPPImage = 32;

            _BPPAlpha = 8;
            _BPPRed = 8;
            _BPPGreen = 8;
            _BPPBlue = 8;
            _MaskAlpha = 0x000000FF;
            _MaskRed = 0xFF000000;
            _MaskGreen = 0x00FF0000;
            _MaskBlue = 0x0000FF00;
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            byte[] paletteBytes = ReadBytes(_PaletteOffset, _PaletteLength);

            Color[] palette = new Color[256];

            for (int i = 0; i < palette.Length; i++)
            {
                int baseByteNum = i * 4;
                int alpha = paletteBytes[baseByteNum + 3];
                int red = paletteBytes[baseByteNum];
                int green = paletteBytes[baseByteNum + 1];
                int blue = paletteBytes[baseByteNum + 2];
                palette[i] = Color.FromArgb(alpha, red, green, blue);
            }

            byte[] imageBytes = ReadBytes(_ImageOffset, Width * Height);
            imageBytes = BLPS2.PS2ImageData.UnSwizzleImageBytes(imageBytes, Width, Height, 8);  // 8 = 8BPP for the main image data

            Bitmap image = CreateBitmap(Width, Height, Stride, PixelFormat.Format8bppIndexed, imageBytes);
            image.Palette = BLI.ColourPalette.FromArray(palette);

            return image;
        }
    }
}
