// BenLincoln.Utility - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.Utility.

// BenLincoln.Utility is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.Utility is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.Utility (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.Utility
{
    public class PerformanceGatherer
    {
        public static PerformanceCounter[] GetPerformanceCounterForProcesses(string[] processNames,
            string perfCounterName)
        {
            ArrayList counterList = new ArrayList();

            foreach (string pName in processNames)
            {
                //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: attempting to obtain performance counter '" +
                //    perfCounterName + "' for process '" + pName + "'.");
                try
                {
                    PerformanceCounter pc = new PerformanceCounter("Process", perfCounterName, pName);
                    counterList.Add(pc);
                    //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: obtained counter");
                }
                catch (Exception ex)
                {
                    //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: exception thrown while attempting to obtain performance counter '" + 
                    //    perfCounterName + "' for process '" + pName + "': " + ex.Message);
                    // do nothing
                }
            }

            PerformanceCounter protoCounter = new PerformanceCounter();
            return (PerformanceCounter[])counterList.ToArray(protoCounter.GetType());
        }

        public static PerformanceCounter[] GetPerformanceCounterForProcesses(Process[] processes,
            string perfCounterName, ExecutionPlatform platform)
        {
            ArrayList pidList = new ArrayList();
            foreach (Process p in processes)
            {
                if (!pidList.Contains(p.Id))
                {
                    pidList.Add(p.Id);
                }
            }
            int[] pids = (int[])pidList.ToArray(int.MinValue.GetType());
            return GetPerformanceCounterForProcesses(pids, perfCounterName, platform);
        }

        public static PerformanceCounter[] GetPerformanceCounterForProcesses(int[] parentProcessIDs,
            string perfCounterName, ExecutionPlatform platform)
        {
            string pCounterName = perfCounterName.ToLower();
            ArrayList counterList = new ArrayList();
            PerformanceCounterCategory pcc = new PerformanceCounterCategory("Process");

            string[] instanceNames = pcc.GetInstanceNames();
            foreach (string iName in instanceNames)
            {
                PerformanceCounter locatedCounter = new PerformanceCounter();
                bool foundCounterName = false;
                bool foundMatchingProcessID = false;
                switch (platform)
                {
                    case ExecutionPlatform.Windows:
                        foreach (PerformanceCounter counter in pcc.GetCounters(iName))
                        {
                            string counterName = counter.CounterName.ToLower();
                            if (counterName == pCounterName)
                            {
                                locatedCounter = counter;
                                foundCounterName = true;
                            }
                            if (counterName == "id process")
                            {
                                foreach (int pid in parentProcessIDs)
                                {
                                    if (counter.RawValue == pid)
                                    {
                                        foundMatchingProcessID = true;
                                        break;
                                    }
                                }
                            }
                        }
                        break;
                    case ExecutionPlatform.UnixLike:
                        // thanks for making this completely different, Mono Project!
                        //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: attempting to obtain a process ID from the string '" + iName + "'.");
                        string[] iNameComponents = iName.Split('/');
                        try
                        {
                            int pid = int.Parse(iNameComponents[0]);
                            foreach (int ppid in parentProcessIDs)
                            {
                                if (ppid == pid)
                                {
                                    foundMatchingProcessID = true;
                                    break;
                                }
                            }
                            foreach (PerformanceCounter counter in pcc.GetCounters(iName))
                            {
                                string counterName = counter.CounterName.ToLower();
                                if (counterName == pCounterName)
                                {
                                    locatedCounter = counter;
                                    foundCounterName = true;
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: exception thrown while trying to convert '" + iNameComponents[0].ToString() + "' to an integer: " + ex.Message);
                        }
                        foreach (PerformanceCounter counter in pcc.GetCounters(iName))
                        {
                            string counterName = counter.CounterName.ToLower();
                            //Console.WriteLine("Debug [GetPerformanceCounterForProcesses]: Counter Name '" + counter.CounterName + "', Instance Name '" +
                                //counter.InstanceName + "', Category Name '" + counter.CategoryName + "', Raw Value " +
                                //counter.RawValue.ToString());
                        }
                        break;
                }
                if (foundMatchingProcessID && foundCounterName)
                {
                    counterList.Add(locatedCounter);
                }
            }

            PerformanceCounter protoCounter = new PerformanceCounter();
            return (PerformanceCounter[])counterList.ToArray(protoCounter.GetType());
        }
    }
}
