using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace BloodOmen2ControlStation
{
    public class ERGVariable
    {
        protected string _Category;
        protected string _Name;
        protected Control[] _BoundControls;

        #region Properties

        public string Category
        {
            get
            {
                return _Category;
            }
        }

        public string Name
        {
            get
            {
                return _Name;
            }
        }

        #endregion

        public ERGVariable(string category, string name, Control boundControl)
        {
            _Category = category;
            _Name = name;
            _BoundControls = new Control[1];
            _BoundControls[0] = boundControl;
        }

        public string GetFullName()
        {
            StringBuilder builder = new StringBuilder();
            if (_Category != "")
            {
                builder.Append(_Category + "\\");
            }
            builder.Append(_Name);
            string fullName = builder.ToString();
            Regex specialChars = new Regex("[^a-zA-Z0-9]");
            if (specialChars.Match(fullName).Success)
            {
                fullName = "\"" + fullName + "\"";
            }
            return fullName;
        }

        public virtual string GetValue()
        {
            return "null";
        }

        public void SetValue(string newValue)
        {
            SetValue(newValue, false);
        }

        public virtual void SetValue(string newValue, bool invertValue)
        {
            // overridden in subclasses
            // invertvalue is for handling the "-somevalue" syntax in the default file
        }

        public virtual void UpdateBoundControls()
        {
            // overridden in subclasses
            // changes the state of the bound control to match the state of the variable
        }

        public virtual void GetValueFromBoundControls()
        {
            // overridden in subclasses
            // changes the state of the variable to match the state of the bound control
        }
    }
}
