// The Eye of RAW
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Eye of RAW.

// The Eye of RAW is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Eye of RAW is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Eye of RAW (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;

namespace EyeOfRAW
{
    public class ImageTransformer
    {

    #region Arrays for PS2 swizzling
        
        public static int[] block32 = 
            {
                 0,  1,  4,  5, 16, 17, 20, 21,
                 2,  3,  6,  7, 18, 19, 22, 23,
                 8,  9, 12, 13, 24, 25, 28, 29,
                10, 11, 14, 15, 26, 27, 30, 31
            };

        public static int[] columnWord32 = 
            {
	             0,  1,  4,  5,  8,  9, 12, 13,
	             2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] blockZ32 = 
            {
	             24, 25, 28, 29, 8, 9, 12, 13,
	             26, 27, 30, 31,10, 11, 14, 15,
	             16, 17, 20, 21, 0, 1, 4, 5,
	             18, 19, 22, 23, 2, 3, 6, 7
            };

        public static int[] columnWordZ32 = 
            {
	             0,  1,  4,  5,  8,  9, 12, 13,
	             2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] block16 = 
            {
	             0,  2,  8, 10,
	             1,  3,  9, 11,
	             4,  6, 12, 14,
	             5,  7, 13, 15,
	            16, 18, 24, 26,
	            17, 19, 25, 27,
	            20, 22, 28, 30,
	            21, 23, 29, 31
            };

        public static int[] columnWord16 = 
            {
	             0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
	             2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] columnHalf16 = 
            {
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1,
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1
            };

        public static int[] blockZ16 = 
            {
	            24,  26,  16, 18,
	            25,  27,  17, 19,
                28,  30, 20, 22,
	            29,  31, 21, 23,
	            8, 10, 0, 2,
	            9, 11, 1, 3,
	            12, 14, 4, 6,
	            13, 15, 5, 7
            };

        public static int[] columnWordZ16 = 
            {
	            0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
	            2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] columnHalfZ16 = 
            {
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1,
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1
            };

        public static int[] blockZ16S = 
            {
	            24,  26,  8, 10,
	            25,  27,  9, 11,
	            16,  18, 0, 2,
	            17,  19, 1, 3,
	            28, 30, 12, 14,
	            29, 31, 13, 15,
	            20, 22, 4, 6,
	            21, 23, 5, 7
            };

        public static int[] columnWordZ16S = 
            {
	            0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
	            2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] columnHalfZ16S = 
            {
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1,
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1
            };

        public static int[] block16S = 
            {
	             0,  2, 16, 18,
	             1,  3, 17, 19,
	             8, 10, 24, 26,
	             9, 11, 25, 27,
	             4,  6, 20, 22,
	             5,  7, 21, 23,
	            12, 14, 28, 30,
	            13, 15, 29, 31
            };

        public static int[] columnWord16S = 
            {
	             0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
	             2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };

        public static int[] columnHalf16S = 
            {
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1,
	            0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 1, 1, 1, 1, 1, 1
            };

        public static int[] block8 = 
            {
	             0,  1,  4,  5, 16, 17, 20, 21,
	             2,  3,  6,  7, 18, 19, 22, 23,
	             8,  9, 12, 13, 24, 25, 28, 29,
	            10, 11, 14, 15, 26, 27, 30, 31
            };

        public static int[] block8a = 
            {
	             0,  1,  2, 3,
	             4,  5,  6, 7,
	             8,  9, 10, 11,
	            12, 13, 14, 15,
	            16, 17, 18, 19,
	            20, 21, 22, 23,
	            24, 25, 26, 27,
	            28, 29, 30, 31
            };

        int[][] columnWord8;    //initialized in constructor
        int[][] columnWord8a;    //initialized in constructor

        public static int[] columnByte8 = 
            {
	            0, 0, 0, 0, 0, 0, 0, 0,  2, 2, 2, 2, 2, 2, 2, 2,
	            0, 0, 0, 0, 0, 0, 0, 0,  2, 2, 2, 2, 2, 2, 2, 2,

	            1, 1, 1, 1, 1, 1, 1, 1,  3, 3, 3, 3, 3, 3, 3, 3,
	            1, 1, 1, 1, 1, 1, 1, 1,  3, 3, 3, 3, 3, 3, 3, 3
            };

        public static int[] block4 = 
            {
	             0,  2,  8, 10,
	             1,  3,  9, 11,
	             4,  6, 12, 14,
	             5,  7, 13, 15,
	            16, 18, 24, 26,
	            17, 19, 25, 27,
	            20, 22, 28, 30,
	            21, 23, 29, 31
            };

        public static int[][] columnWord4;    //initialized in constructor

        public static int[] columnByte4 = 
            {
                0, 0, 0, 0, 0, 0, 0, 0,  2, 2, 2, 2, 2, 2, 2, 2,  4, 4, 4, 4, 4, 4, 4, 4,  6, 6, 6, 6, 6, 6, 6, 6,
                0, 0, 0, 0, 0, 0, 0, 0,  2, 2, 2, 2, 2, 2, 2, 2,  4, 4, 4, 4, 4, 4, 4, 4,  6, 6, 6, 6, 6, 6, 6, 6,

                1, 1, 1, 1, 1, 1, 1, 1,  3, 3, 3, 3, 3, 3, 3, 3,  5, 5, 5, 5, 5, 5, 5, 5,  7, 7, 7, 7, 7, 7, 7, 7,
                1, 1, 1, 1, 1, 1, 1, 1,  3, 3, 3, 3, 3, 3, 3, 3,  5, 5, 5, 5, 5, 5, 5, 5,  7, 7, 7, 7, 7, 7, 7, 7
            };

		#endregion

        public ImageTransformer()
        {
            //initialize multi-dimensional arrays
            columnWord8 = new int[4][];
            columnWord8[0] = new int[64] 
            {
		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,

		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7
            };
            columnWord8[1] = new int[64] 
            {
		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,

		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };
            columnWord8[2] = new int[64] 
            {
		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,

		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7
            };
            columnWord8[3] = new int[64] 
            {
		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,

		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };

            columnWord4 = new int[2][];
            columnWord4[0] = new int[128] 
            {
		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,

		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7
            };
            columnWord4[1] = new int[128] 
            {
		         8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,   8,  9, 12, 13,  0,  1,  4,  5,
		        10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,  10, 11, 14, 15,  2,  3,  6,  7,

		         0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,   0,  1,  4,  5,  8,  9, 12, 13,
		         2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15,   2,  3,  6,  7, 10, 11, 14, 15
            };
        }

        public byte[] ReorderByteArray(byte[] inData)
        {
            byte[] outData;
            int dataLength = inData.GetUpperBound(0) + 1;
            outData = new byte[dataLength];

            int currentInPos = 0;
            int currentOutPos = 0;

            int numBlocks = dataLength / 256;

            for (int i = 0; i < numBlocks; i++)
            {
                int blockX = (i % 8);
                int blockY = (i / 8);
                byte[] currentBlock = new byte[256];
                //for (int b = 0; b < 256; b++)
                //{
                //    currentBlock[b] = inData[(i * 256) + b];
                //}
                for (int y = 0; y < 16; y++)
                {
                    for (int x = 0; x < 16; x++)
                    {
                        currentBlock[x + (y * 16)] = inData[(blockY * 8 * 256) + (blockX * 16) + (y * 8 * 256) + x];
                    }
                }
                for (int y = 0; y < 16; y++)
                {
                    for (int x = 0; x < 16; x++)
                    {
                        //outData[currentOutPos] = currentBlock[columnWord8[j][k] + (j * 64)];
                        outData[currentOutPos] = currentBlock[x + (y * 16)];
                        currentOutPos++;
                    }
                }
            }

            //int numBlocks = dataLength / 256;

            //int[] clutIndex = BuildIDTex8ClutIndex();

            //for (int i = 0; i < numBlocks; i++)
            //{
            //    byte[] currentBlock = new byte[256];
            //    for (int b = 0; b < 256; b++)
            //    {
            //        currentBlock[b] = inData[(i * 256) + b];
            //    }
            //    for (int j = 0; j < 256; j++)
            //    {
            //        //outData[currentOutPos] = currentBlock[columnWord8[j][k] + (j * 64)];
            //        outData[currentOutPos] = currentBlock[clutIndex[j]];
            //        currentOutPos++;
            //     }
            //}

            //int numBlocks = dataLength / 32;

            //for (int i = 0; i < numBlocks; i++)
            //{
            //    byte[] currentBlock = new byte[32];
            //    for (int b = 0; b < 32; b++)
            //    {
            //        currentBlock[b] = inData[(i * 32) + b];
            //    }
            //    for (int j = 0; j < 32; j++)
            //    {
            //        //outData[currentOutPos] = currentBlock[columnWord8[j][k] + (j * 64)];
            //        outData[currentOutPos] = currentBlock[columnWordZ16S[j] + columnHalfZ16S[j]];
            //        currentOutPos++;
            //    }
            //}

            //int numBlocks = dataLength / 64;

            //for (int i = 0; i < numBlocks; i++)
            //{
            //    byte[] currentBlock = new byte[64];
            //    for (int b = 0; b < 64; b++)
            //    {
            //        currentBlock[b] = inData[(i * 64) + b];
            //    }
            //    for (int j = 0; j < 64; j++)
            //    {
            //        //outData[currentOutPos] = currentBlock[columnWord8[j][k] + (j * 64)];
            //        outData[currentOutPos] = currentBlock[columnWord8[1][j] + columnByte8[j]];
            //        currentOutPos++;
            //    }
            //}


            return outData;
        }

        protected int[] BuildIDTex8ClutIndex()
        {
            int[] clutIndex = new int[256];
            int currentIndex = 0;
            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 16 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 8 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 24 + j] = currentIndex;
                    currentIndex++;
                }
            }
            return clutIndex;
        }

        public byte[] XReadPSTwo32BitImage(byte[] inData, int width, int height, int dbw, int[] blockTable, int[] columnTable)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int currentOutPos = 0;
            
            for (int i = 0; i < outData.GetUpperBound(0); i++)
            {
                int currentPage = i / 8192;
                int pageRemainder = i - currentPage * 8192;
                int currentBlock = pageRemainder / 256;
                int blockRemainder = pageRemainder - currentBlock * 256;
                int currentColumn = blockRemainder / 64;
                int columnRemainder = blockRemainder - currentColumn * 64;
                int currentPixel = columnRemainder / 4;
                int pixelRemainder = columnRemainder - (currentPixel * 4);

                int blockIndex = blockTable[currentBlock];
                int columnIndex = columnTable[currentPixel] + pixelRemainder + (currentColumn * 64);

                int currentInPos = (currentPage * 8192) + (blockIndex * 256) + (columnIndex);
                outData[currentOutPos] = inData[currentInPos];
                currentOutPos++;
            }

            return outData;
        }

        public byte[] XReadPSMZ32(byte[] inData, int width, int height, int dbw)
        {
            return XReadPSTwo32BitImage(inData, width, height, dbw, blockZ32, columnWordZ32);
        }


        public byte[] ReadPSTwo32BitImage(byte[] inData, int width, int height, int dbw, int[] blockTable, int[] columnTable)
        {
            uint[] inDataUInt = ByteArrayToUIntArray(inData, false);
            uint[] outDataUInt = new uint[inDataUInt.GetUpperBound(0) + 1];

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    int pageX = x / 64;
                    int pageY = y / 32;
                    int page = pageX + pageY * dbw;

                    int px = x - (pageX * 64);
                    int py = y - (pageY * 32);

                    int blockX = px / 8;
                    int blockY = py / 8;
                    int block = blockTable[blockX + blockY * 8];

                    int bx = px - blockX * 8;
                    int by = py - blockY * 8;

                    int column = by / 2;

                    int cx = bx;
                    int cy = by - column * 2;
                    int cw = columnTable[cx + cy * 8];

                    currentInPos = (page * 2048) + (block * 64) + (column * 16) + cw;
                    outDataUInt[currentOutPos] = inDataUInt[currentInPos];
                    //outDataUInt[currentInPos] = inDataUInt[currentOutPos];
                    currentOutPos++;
                }
            }

            return UIntArrayToByteArray(outDataUInt);
        }

        public byte[] ReadTexPSMCT32(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo32BitImage(inData, width, height, dbw, block32, columnWord32);
        }

        public byte[] ReadTexPSMZ32(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo32BitImage(inData, width, height, dbw, blockZ32, columnWordZ32);
        }

        public byte[] ReadPSTwo16BitImage(byte[] inData, int width, int height, int dbw, int[] blockTable, int[] columnWordTable, int[] columnHalfTable)
        {
            ushort[] inDataUShort = ByteArrayToUShortArray(inData, false);
            ushort[] outDataUShort = new ushort[inDataUShort.GetUpperBound(0) + 1];

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    int pageX = x / 64;
                    int pageY = y / 64;
                    int page = pageX + pageY * dbw;

                    int px = x - (pageX * 64);
                    int py = y - (pageY * 64);

                    int blockX = px / 16;
                    int blockY = py / 8;
                    int block = blockTable[blockX + blockY * 4];

                    int bx = px - blockX * 16;
                    int by = py - blockY * 8;

                    int column = by / 2;

                    int cx = bx;
                    int cy = by - column * 2;
                    int cw = columnWordTable[cx + cy * 16];
                    int ch = columnHalfTable[cx + cy * 16];

                    currentInPos = (page * 2048) + (block * 64) + (column * 16) + cw;
                    currentInPos *= 2;
                    currentInPos += ch;
                    outDataUShort[currentOutPos] = inDataUShort[currentInPos];
                    //outDataUShort[currentInPos] = inDataUShort[currentOutPos];
                    currentOutPos++;
                }
            }

            return UShortArrayToByteArray(outDataUShort);
        }

        public byte[] ReadTexPSMCT16(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo16BitImage(inData, width, height, dbw, block16, columnWord16, columnHalf16);
        }

        public byte[] ReadTexPSMCT16S(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo16BitImage(inData, width, height, dbw, block16S, columnWord16S, columnHalf16S);
        }

        public byte[] ReadTexPSMZ16(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo16BitImage(inData, width, height, dbw, blockZ16, columnWordZ16, columnHalfZ16);
        }

        public byte[] ReadTexPSMZ16S(byte[] inData, int width, int height, int dbw)
        {
            return ReadPSTwo16BitImage(inData, width, height, dbw, blockZ16S, columnWordZ16S, columnHalfZ16S);
        }
  
        public byte[] ReadTexPSMT8(byte[] inData, int width, int height, int dbw)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int currentInPos = 0;
            int currentOutPos = 0;
            //int maxInPos = 0;

            //for (int i = 0; i < outData.GetUpperBound(0); i++)
            //{
            //    int currentPage = i / 8192;
            //    int pageRemainder = i - currentPage * 8192;
            //    int currentBlock = pageRemainder / 256;
            //    int blockRemainder = pageRemainder - currentBlock * 256;
            //    int currentColumn = blockRemainder / 64;
            //    int columnRemainder = blockRemainder - currentColumn * 64;

            //    int blockIndex = block8[currentBlock];
            //    int columnIndex = columnWord8[currentColumn][columnRemainder] + columnByte8[columnRemainder] + (currentColumn * 64);

            //    int currentInPos = (currentPage * 8192) + (blockIndex * 256) + (columnIndex);
            //    outData[currentOutPos] = inData[currentInPos];
            //    currentOutPos++;
            //}

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    //page = how many pages we're in to the file
                    int pageX = x / 128;
                    int pageY = y / 64;
                    int page = pageX + (pageY * dbw);

                    //px = remainder inside the current page
                    int px = x - (pageX * 128);
                    int py = y - (pageY * 64);

                    //block = how many blocks we're into the page
                    int blockX = px / 16;
                    int blockY = py / 16;
                    int block = block8[blockX + (blockY * 8)];

                    //bx = remainder inside the current block
                    int bx = px - blockX * 16;
                    int by = py - blockY * 16;

                    //column = column#
                    int column = by / 4;

                    //figure out how many dwords into the column we are
                    int cx = bx;
                    int cy = by - column * 4;
                    int cw = columnWord8[column][cx + (cy * 16)];
                    int cb = columnByte8[cx + (cy * 16)];

                    currentInPos = ((page * 2048) + (block * 64) + (column * 16) + cw) * 4;
                    currentInPos += cb;
                    //outData[currentOutPos + ch] = inData[currentInPos];
                    outData[currentOutPos] = inData[currentInPos];
                    //outData[currentInPos] = inData[currentOutPos];
                    currentOutPos++;
                    //maxInPos = Math.Max(maxInPos, currentInPos);
                }
            }
            return outData;
        }

        public byte[] ReadTexPSMT4(byte[] inData, int width, int height, int dbw)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];
            for (int i = 0; i < outData.GetUpperBound(0); i++)
            {
                outData[i] = (byte)0;
            }

            //int currentInPos = 0;
            int currentOutPos = 0;

            for (int i = 0; i < outData.GetUpperBound(0); i++)
            {
                int currentPage = i / 8192;
                int pageRemainder = i - currentPage * 8192;
                int currentBlock = pageRemainder / 256;
                int blockRemainder = pageRemainder - currentBlock * 256;
                int currentColumn = blockRemainder / 64;
                int columnRemainder = blockRemainder - currentColumn * 64;

                int blockIndex = block4[currentBlock];
                int columnIndex = columnWord4[currentColumn % 2][columnRemainder] + columnByte4[columnRemainder] + (currentColumn * 64);

                int currentInPos = (currentPage * 8192) + (blockIndex * 256) + (columnIndex);
                outData[currentOutPos] = inData[currentInPos];
                currentOutPos++;
            }

            //bool odd = false;

            //for (int y = 0; y < height; y++)
            //{
            //    for (int x = 0; x < width; x++)
            //    {
            //        int pageX = x / 128;
            //        int pageY = y / 128;
            //        int page = pageX + pageY * dbw;

            //        int px = x - (pageX * 128);
            //        int py = y - (pageY * 128);

            //        int blockX = px / 32;
            //        int blockY = py / 16;
            //        int block = block4[blockX + blockY * 4];

            //        int bx = px - blockX * 32;
            //        int by = py - blockY * 16;

            //        int column = by / 4;

            //        int cx = bx;
            //        int cy = by - column * 4;
            //        int cw = columnWord4[column & 1][cx + cy * 32];
            //        int cb = columnByte4[cx + cy * 32];

            //        currentInPos = ((page * 2048) + (block * 64) + (column * 16) + cw) * 4;
            //        currentInPos += cb;
            //        //outData[currentOutPos + ch] = inData[currentInPos];
            //        if ((cb & 1) == 1)
            //        {
            //            if (odd)
            //            {
            //                outData[currentOutPos] = (byte)((outData[currentOutPos] & 0x0f) | (inData[currentInPos + (cb >> 1)] & 0xf0));
            //            }
            //            else
            //            {
            //                outData[currentOutPos] = (byte)((outData[currentOutPos] & 0xf0) | ((inData[currentInPos + (cb >> 1)] >> 4) & 0x0f));
            //            }
            //        }
            //        else
            //        {
            //            if (odd)
            //            {
            //                outData[currentOutPos] = (byte)((outData[currentOutPos] & 0x0f) | ((inData[currentInPos + (cb >> 1)] << 4) & 0xf0));
            //            }
            //            else
            //            {
            //                outData[currentOutPos] = (byte)((outData[currentOutPos] & 0xf0) | (inData[currentInPos + (cb >> 1)] & 0x0f));
            //            }
            //        }

            //        if (odd)
            //        {
            //            currentOutPos++;
            //        }
            //        odd = !odd;
            //    }
            //}
            return outData;
        }

        public ushort[] ByteArrayToUShortArray(byte[] inData, bool convertEndian)
        {
            int dataLength = inData.GetUpperBound(0) + 1;
            if ((dataLength % 2) == 1)
            {
                throw new Exception("Can't convert a byte array to a ushort array unless its length is divisible by 2.");
            }
            ushort[] outData = new ushort[dataLength / 2];

            for (int i = 0; i <= outData.GetUpperBound(0); i++)
            {
                ushort byte1 = (ushort)inData[i * 2];
                ushort byte2 = (ushort)inData[(i * 2) + 1];
                if (convertEndian)
                {
                    outData[i] = (ushort)((byte2 << 8) | byte1);
                }
                else
                {
                    outData[i] = (ushort)((byte1 << 8) | byte2);
                }
            }

            return outData;
        }

        public uint[] ByteArrayToUIntArray(byte[] inData, bool convertEndian)
        {
            int dataLength = inData.GetUpperBound(0) + 1;
            if ((dataLength % 4) > 0)
            {
                throw new Exception("Can't convert a byte array to a uint array unless its length is divisible by 4.");
            }
            uint[] outData = new uint[dataLength / 4];

            for (int i = 0; i <= outData.GetUpperBound(0); i++)
            {
                uint byte1 = (uint)inData[i * 4];
                uint byte2 = (uint)inData[(i * 4) + 1];
                uint byte3 = (uint)inData[(i * 4) + 2];
                uint byte4 = (uint)inData[(i * 4) + 3];
                if (convertEndian)
                {
                    outData[i] = (uint)((byte4 << 24) | (byte3 << 16) | (byte2 << 8) | byte1);
                }
                else
                {
                    outData[i] = (uint)((byte1 << 24) | (byte2 << 16) | (byte3 << 8) | byte4);
                }
            }

            return outData;
        }

        public byte[] UIntArrayToByteArray(uint[] inData)
        {
            int dataLength = inData.GetUpperBound(0) + 1;
            byte[] outData = new byte[dataLength * 4];

            int outCounter = 0;
            for (int i = 0; i < dataLength; i++)
            {
                uint inVal = (uint)inData[i];
                uint byte1 = inVal;
                uint byte2 = inVal;
                uint byte3 = inVal;
                uint byte4 = inVal;
                byte1 = (uint)((byte1 & 0xFF000000) >> 24);
                byte2 = (uint)((byte2 & 0xFF0000) >> 16);
                byte3 = (uint)((byte3 & 0xFF00) >> 8);
                byte4 &= 0xFF;
                outData[outCounter] = (byte)byte4;
                outData[outCounter + 1] = (byte)byte3;
                outData[outCounter + 2] = (byte)byte2;
                outData[outCounter + 3] = (byte)byte1;
                outCounter += 4;
            }

            return outData;
        }

        public byte[] UShortArrayToByteArray(ushort[] inData)
        {
            int dataLength = inData.GetUpperBound(0) + 1;
            byte[] outData = new byte[dataLength * 2];

            int outCounter = 0;
            for (int i = 0; i < dataLength; i++)
            {
                ushort inVal = (ushort)inData[i];
                ushort byte1 = inVal;
                ushort byte2 = inVal;
                byte1 = (ushort)((byte1 & 0xFF00) >> 8);
                byte2 &= 0xFF;
                outData[outCounter] = (byte)byte2;
                outData[outCounter + 1] = (byte)byte1;
                outCounter += 2;
            }

            return outData;
        }

    }
}
