// The Eye of RAW
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Eye of RAW.

// The Eye of RAW is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Eye of RAW is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Eye of RAW (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;

namespace EyeOfRAW
{
    public class ImageFile
    {
        protected string mFileName;
        protected string mFilePath;
        protected string mTypeDescription;

        protected long mFileSize;
        protected long mImageDataOffset;
        //protected ushort mFileType;
        //protected ushort mImageType;
        protected ImageTypes mImageType;
        protected FileTypes mFileType;

        protected uint mDataLength;
        protected uint mBytesPerLine;

        protected int mImageWidth;
        protected int mImageHeight;

        protected int mBPPImage;
        protected int mBPPAlpha;
        protected int mBPPRed;
        protected int mBPPGreen;
        protected int mBPPBlue;

        protected uint mMaskAlpha;
        protected uint mMaskRed;
        protected uint mMaskGreen;
        protected uint mMaskBlue;

        protected int mStartBitAlpha;
        protected int mStartBitRed;
        protected int mStartBitGreen;
        protected int mStartBitBlue;

        protected int mSubImageCount;
        protected int mSubHorizontalCount;
        protected int mSubVerticalCount;
        protected int[] mSubImageWidths;

        protected Bitmap mImage;

        #region Constants

        //public const ushort IMAGETYPE_GREY = 0;
        //public const ushort IMAGETYPE_RGB_PACKED = 2;

        public enum ImageTypes
        {
            Grey,
            Indexed,
            RGBPacked
        }

        public enum FileTypes
        {
            SRTex,
            RAW_Undetermined,
            SR2_PS2_AirForge,
            SR2_PS2_AirForge_ARGB,
            SR2_PS2_AirForge_Indexed,
            SR2_PS2_ARGB,
            SR2_PS2_Indexed,
            SR2_PC,
            Defiance_PS2_ARGB,
            Defiance_PS2_Indexed,
            Defiance_PC,
            TRL_PS2,
            TRL_PSP_ARGB,
            TRL_PSP_Indexed,
            Unknown
        }

        //public const ushort FILETYPE_SRTEX = 100;
        //public const ushort FILETYPE_RAW_UNDETERMINED = 125;
        //public const ushort FILETYPE_SR2_PS2_AIRFORGE = 130;
        //public const ushort FILETYPE_SR2_PS2_ARGB = 150;
        //public const ushort FILETYPE_SR2_PS2_INDEXED = 160;
        //public const ushort FILETYPE_SR2_PC = 200;
        //public const ushort FILETYPE_DEFIANCE_PS2 = 250;
        //public const ushort FILETYPE_DEFIANCE_PC = 300;
        //public const ushort FILETYPE_TRL_PS2 = 350;
        //public const ushort FILETYPE_TRL_PSP = 400;
        //public const ushort FILETYPE_UNKNOWN = 4096;

        public const string FILEDESC_SRTEX = "Soul Reaver (Playstation) Texture Array";
        public const string FILEDESC_SR2_PS2_AIRFORGE_ARGB = "Soul Reaver 2 (Playstation 2) Air Forge Demo RAW Bitmap";
        public const string FILEDESC_SR2_PS2_AIRFORGE_INDEXED = "Soul Reaver 2 (Playstation 2) Air Forge Demo RAW Indexed Image";
        public const string FILEDESC_SR2_PS2_ARGB = "Soul Reaver 2 (Playstation 2) RAW Bitmap";
        public const string FILEDESC_SR2_PS2_INDEXED = "Soul Reaver 2 (Playstation 2) RAW Indexed Image";
        public const string FILEDESC_SR2_PC = "Soul Reaver 2 (PC) RAW Bitmap";
        public const string FILEDESC_DEFIANCE_PS2_ARGB = "Defiance / TR:Legend (Playstation 2) RAW Bitmap";
        public const string FILEDESC_DEFIANCE_PS2_INDEXED = "Defiance / TR:Legend (Playstation 2) RAW Indexed Image";
        public const string FILEDESC_DEFIANCE_PC = "Defiance (PC) RAW Bitmap";
        public const string FILEDESC_TRL_PS2 = "TR:Legend (Playstation 2) RAW Bitmap";
        public const string FILEDESC_TRL_PSP_ARGB = "TR:Legend (PSP) RAW Bitmap";
        public const string FILEDESC_TRL_PSP_INDEXED= "TR:Legend (PSP) RAW Indexed Image";
        public const string FILEDESC_UNKNOWN = "Unknown";

        public const long OFFSET_RAW_HEADER = 8;


        #endregion

        #region Properties

        public string FileName
        {
            get
            {
                return mFileName;
            }
            set
            {
                mFileName = value;
            }
        }

        public string FilePath
        {
            get
            {
                return mFilePath;
            }
            set
            {
                mFilePath = value;
            }
        }

        public string TypeDescription
        {
            get
            {
                return mTypeDescription;
            }
            set
            {
                mTypeDescription = value;
            }
        }

        public long FileSize
        {
            get
            {
                return mFileSize;
            }
            set
            {
                mFileSize = value;
            }
        }

        public long ImageDataOffset
        {
            get
            {
                return mImageDataOffset;
            }
            set
            {
                mImageDataOffset = value;
            }
        }

        public FileTypes FileType
        {
            get
            {
                return mFileType;
            }
            set
            {
                mFileType = value;
            }
        }

        public ImageTypes ImageType
        {
            get
            {
                return mImageType;
            }
            set
            {
                mImageType = value;
            }
        }

        public uint DataLength
        {
            get
            {
                return mDataLength;
            }
            set
            {
                mDataLength = value;
            }
        }

        public uint BytesPerLine
        {
            get
            {
                return mBytesPerLine;
            }
            set
            {
                mBytesPerLine = value;
            }
        }

        public int ImageWidth
        {
            get
            {
                return mImageWidth;
            }
            set
            {
                mImageWidth = value;
            }
        }

        public int ImageHeight
        {
            get
            {
                return mImageHeight;
            }
            set
            {
                mImageHeight = value;
            }
        }

        public int BPPImage
        {
            get
            {
                return mBPPImage;
            }
            set
            {
                mBPPImage = value;
            }
        }

        public int BPPAlpha
        {
            get
            {
                return mBPPAlpha;
            }
            set
            {
                mBPPAlpha = value;
            }
        }

        public int BPPRed
        {
            get
            {
                return mBPPRed;
            }
            set
            {
                mBPPRed = value;
            }
        }

        public int BPPGreen
        {
            get
            {
                return mBPPGreen;
            }
            set
            {
                mBPPGreen = value;
            }
        }

        public int BPPBlue
        {
            get
            {
                return mBPPBlue;
            }
            set
            {
                mBPPBlue = value;
            }
        }

        public uint MaskAlpha
        {
            get
            {
                return mMaskAlpha;
            }
            set
            {
                mMaskAlpha = value;
            }
        }

        public uint MaskRed
        {
            get
            {
                return mMaskRed;
            }
            set
            {
                mMaskRed = value;
            }
        }

        public uint MaskGreen
        {
            get
            {
                return mMaskGreen;
            }
            set
            {
                mMaskGreen = value;
            }
        }

        public uint MaskBlue
        {
            get
            {
                return mMaskBlue;
            }
            set
            {
                mMaskBlue = value;
            }
        }

        public int StartBitAlpha
        {
            get
            {
                return mStartBitAlpha;
            }
            set
            {
                mStartBitAlpha = value;
            }
        }

        public int StartBitRed
        {
            get
            {
                return mStartBitRed;
            }
            set
            {
                mStartBitRed = value;
            }
        }

        public int StartBitGreen
        {
            get
            {
                return mStartBitGreen;
            }
            set
            {
                mStartBitGreen = value;
            }
        }

        public int StartBitBlue
        {
            get
            {
                return mStartBitBlue;
            }
            set
            {
                mStartBitBlue = value;
            }
        }

        public int[] SubImageWidths
        {
            get
            {
                return mSubImageWidths;
            }
            set
            {
                mSubImageWidths = value;
            }
        }

        public int SubImageCount
        {
            get
            {
                return mSubImageCount;
            }
            set
            {
                mSubImageCount = value;
            }
        }

        public int SubHorizontalCount
        {
            get
            {
                return mSubHorizontalCount;
            }
            set
            {
                mSubHorizontalCount = value;
            }
        }

        public int SubVerticalCount
        {
            get
            {
                return mSubVerticalCount;
            }
            set
            {
                mSubVerticalCount = value;
            }
        }

        public Bitmap Image
        {
            get
            {
                return mImage;
            }
            set
            {
                mImage = value;
            }
        }


        #endregion

        public ImageFile(string filePath)
        {
            mFilePath = filePath;
            int lastBackSlash = mFilePath.LastIndexOf('\\') + 1;
            mFileName = mFilePath.Substring(lastBackSlash, mFilePath.Length - lastBackSlash);

            if (!File.Exists(mFilePath))
            {
                throw new FileNotFoundException("Can't find file " + mFilePath);
            }
            FileInfo fInfo;
            fInfo = new FileInfo(mFilePath);
            mFileSize = fInfo.Length;

            GetHeaderData();

            if (mFileType != FileTypes.SRTex)
            {
                GetRAWParameters();
            }
            else
            {
                GetSRTexParameters();
            }
        }

        protected void GetHeaderData()
        {
            FileStream iStream;
            BinaryReader iReader;

            uint header1, header2;

            iStream = new FileStream(mFilePath, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);
            header1 = iReader.ReadUInt32();
            header2 = iReader.ReadUInt32();
            iReader.Close();
            iStream.Close();
            if ((header1 == 1380013857))// && (header2 == 128))
            {
                if (header2 == 84)
                {
                    mFileType = FileTypes.SR2_PS2_AirForge;
                }
                else
                {
                    mFileType = FileTypes.RAW_Undetermined;
                }
            }
            else
            {
                //the SR texture data doesn't have a header, so we have to try 
                //anything without a !WAR header as one
                mFileType = FileTypes.SRTex;
            }
        }

        protected void GetRAWParameters()
        {
            FileStream iStream;
            BinaryReader iReader;
            iStream = new FileStream(mFilePath, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);
            iStream.Seek(OFFSET_RAW_HEADER, SeekOrigin.Begin);
            
            //get all the parameters, even if they're not really in each type of file
            uint tDataLength = iReader.ReadUInt32();
            uint tFileType = iReader.ReadUInt32();
            uint tImageType = iReader.ReadUInt32();
            uint tWidth = iReader.ReadUInt32();
            uint tHeight = iReader.ReadUInt32();
            uint tBytesPerLine = iReader.ReadUInt32();
            uint tBPP = iReader.ReadUInt32();
            uint tBPP1 = iReader.ReadUInt32();
            uint tBPP2 = iReader.ReadUInt32();
            uint tBPP3 = iReader.ReadUInt32();
            uint tBPP4 = iReader.ReadUInt32();
            uint tMask1 = ReadMask(iReader);
            uint tMask2 = ReadMask(iReader);
            uint tMask3 = ReadMask(iReader);
            uint tMask4 = ReadMask(iReader);
            uint tUnknown1 = iReader.ReadUInt32();
            uint tUnknown2 = iReader.ReadUInt32();
            uint tUnknown3 = iReader.ReadUInt32();
            uint tUnknown4 = iReader.ReadUInt32();
            uint[] tSubWidths = new uint[6];
            for (int i = 0; i <= tSubWidths.GetUpperBound(0); i++)
            {
                tSubWidths[i] = iReader.ReadUInt32();
            }
            uint tUnknown5 = iReader.ReadUInt32();
            uint tUnknown6 = iReader.ReadUInt32();
            uint tUnknown7 = iReader.ReadUInt32();

            iReader.Close();
            iStream.Close();

            //transform the bitmasks if necessary
            if (tBPP == 16)
            {
                tMask1 = TransformMask(tMask1);
                tMask2 = TransformMask(tMask2);
                tMask3 = TransformMask(tMask3);
                tMask4 = TransformMask(tMask4);
            }

            //get shared parameters
            mImageWidth = (int)tWidth;
            mImageHeight = (int)tHeight;
            mBPPImage = (int)tBPP;
            mDataLength = tDataLength;
            mTypeDescription = "";
            mSubImageCount = 0;

            //figure out what the file type is
            if (mFileType == FileTypes.SR2_PS2_AirForge)
            {
                if ((tFileType == 4) && (tImageType == 4194304))
                {
                    mFileType = FileTypes.SR2_PS2_AirForge_Indexed;
                }
                else
                {
                    mFileType = FileTypes.SR2_PS2_AirForge_ARGB;
                }
            }
            else if ((tFileType == 16) && (tImageType == 16))
            {
                mFileType = FileTypes.Defiance_PC;
            }
            else if (((tFileType == 4)) && (tImageType == 4194304))
            {
                mFileType = FileTypes.SR2_PS2_Indexed;   //16-bit
            }
            //this next set of nasty hacks is necessary because there's no proper way to determine the difference
            //between files from the PS2 and PC versions of SR2 just by looking at their headers
            //even though the file data is DEFINITELY incompatible
            else if (((tFileType == 3)) && (tImageType == 131072) && (tSubWidths[0] == 1024) && (tUnknown3 != 7))
            {
                mFileType = FileTypes.SR2_PS2_ARGB;   //32-bit
            }
            //else if ((tFileType == 12) && (tImageType == 512))
            //{
            //    mFileType = FILETYPE_TRL_PS2;
            //}
            else if (((tFileType == 1) || (tFileType == 2)) && (tBPP == 16))
            {
                mFileType = FileTypes.TRL_PSP_ARGB;
            }
            else if (((tFileType == 12) || (tFileType == 512)) && (tBPP == 8) && (tSubWidths[0] > 1023))
            {
                mFileType = FileTypes.TRL_PSP_Indexed;
            }
            else if (((tFileType == 3) || (tFileType == 0)) &&
               ((tImageType == 16) || (tImageType == 112)))
            {
                mFileType = FileTypes.Defiance_PS2_ARGB;
            }
            else if ((tFileType == 12) && (tImageType == 512))
            {
                mFileType = FileTypes.Defiance_PS2_Indexed;
            }
            else if ((((tFileType == 3) || (tFileType == 0)) &&
               (tImageType != 16) && (tImageType != 112)))
            {
                mFileType = FileTypes.SR2_PC;
            }
            else
            {
                mFileType = FileTypes.Unknown;
                throw new UnrecognizedFileTypeException("Unrecognized RAW file type");
            }

            //set the colour information based on the type of file
            if ((mFileType == FileTypes.SR2_PS2_ARGB))
            {
                mImageDataOffset = 144;
                mBPPAlpha = (int)tBPP1;
                mBPPRed = (int)tBPP2;
                mBPPGreen = (int)tBPP3;
                mBPPBlue = (int)tBPP4;
                mMaskAlpha = tMask1;
                mMaskRed = tMask2;
                mMaskGreen = tMask3;
                mMaskBlue = tMask4;
                mTypeDescription = "Uncompressed Tiled RGB Image";
            }
            if ((mFileType == FileTypes.SR2_PS2_Indexed))
            {
                mImageDataOffset = 144;
                mBPPAlpha = (int)tBPP1;
                mBPPRed = (int)tBPP2;
                mBPPGreen = (int)tBPP3;
                mBPPBlue = (int)tBPP4;
                mMaskAlpha = tMask1;
                mMaskRed = tMask2;
                mMaskGreen = tMask3;
                mMaskBlue = tMask4;
                mTypeDescription = "Uncompressed Tiled Indexed Image";
            }
            if ((mFileType == FileTypes.SR2_PC))
            {
                mImageDataOffset = 128;
                if (mBPPImage == 32)
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP4;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP2;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask4;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask2;
                }
                else
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP2;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP4;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask2;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask4;
                }
                mTypeDescription = "Uncompressed ARGB Image";
            }
            if ((mFileType == FileTypes.SR2_PS2_AirForge_ARGB))
            {
                mImageDataOffset = 84;  //128
                if (mBPPImage == 32)
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP2;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP4;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask2;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask4;
                }
                else
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP2;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP4;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask2;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask4;
                }
                mTypeDescription = "Uncompressed ARGB Image";
            }

            if ((mFileType == FileTypes.SR2_PS2_AirForge_Indexed))
            {
                mImageDataOffset = 84;
                if (mBPPImage == 32)
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP2;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP4;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask2;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask4;
                }
                else
                {
                    mBPPAlpha = (int)tBPP1;
                    mBPPRed = (int)tBPP2;
                    mBPPGreen = (int)tBPP3;
                    mBPPBlue = (int)tBPP4;
                    mMaskAlpha = tMask1;
                    mMaskRed = tMask2;
                    mMaskGreen = tMask3;
                    mMaskBlue = tMask4;
                }
                mTypeDescription = "Uncompressed Indexed Image";
            }

            if ((mFileType == FileTypes.Defiance_PS2_ARGB))
            {
                mImageDataOffset = 144;
                //fix image width if it's not a multiple of 64
                if ((mImageWidth % 64) > 0)
                {
                    mImageWidth += (64 - (mImageWidth % 64));
                }
                mBPPAlpha = (int)tBPP1;
                mBPPRed = (int)tBPP2;
                mBPPGreen = (int)tBPP3;
                mBPPBlue = (int)tBPP4;
                mMaskAlpha = tMask1;
                mMaskRed = tMask2;
                mMaskGreen = tMask3;
                mMaskBlue = tMask4;
                //copy in the sub image data
                mSubImageWidths = new int[6];
                for (int i = 0; i <= tSubWidths.GetUpperBound(0); i++)
                {
                    mSubImageWidths[i] = (int)tSubWidths[i];
                }
                //figure out how many sub-images there are
                for (int i = 0; i <= mSubImageWidths.GetUpperBound(0); i++)
                {
                    if (mSubImageWidths[i] > 0)
                    {
                        mSubImageCount++;
                    }
                }
                switch (mSubImageCount)
                {
                    case 2:
                        mSubHorizontalCount = 2;
                        mSubVerticalCount = 1;
                        break;
                    case 4:
                        mSubHorizontalCount = 2;
                        mSubVerticalCount = 2;
                        break;
                    case 6:
                        mSubHorizontalCount = 3;
                        mSubVerticalCount = 2;
                        break;
                    default:
                        mSubHorizontalCount = mSubImageCount;
                        mSubVerticalCount = 1;
                        break;
                }
                if (mSubImageCount > 1)
                {
                    mTypeDescription = "Uncompressed Tiled ARGB Image";
                }
                else
                {
                    mTypeDescription = "Uncompressed ARGB Image";
                }
            }
            if ((mFileType == FileTypes.Defiance_PS2_Indexed))
            {
                mImageDataOffset = 144;
                //this type lies and claims to be 16-bit when really it's 32
                mBPPImage = 32;
                //manually set all parameters
                mBPPAlpha = 8;
                mBPPRed = 8;
                mBPPGreen = 8;
                mBPPBlue = 8;
                mMaskAlpha = 0x000000FF;
                mMaskRed = 0xFF000000;
                mMaskGreen = 0x00FF0000;
                mMaskBlue = 0x0000FF00;

                ////fix image width if it's not a multiple of 64
                //if ((mImageWidth % 64) > 0)
                //{
                //    mImageWidth += (64 - (mImageWidth % 64));
                //}

                //copy in the sub image data
                mSubImageWidths = new int[6];
                for (int i = 0; i <= tSubWidths.GetUpperBound(0); i++)
                {
                    mSubImageWidths[i] = (int)tSubWidths[i];
                }
                //figure out how many sub-images there are
                for (int i = 0; i <= mSubImageWidths.GetUpperBound(0); i++)
                {
                    if (mSubImageWidths[i] > 0)
                    {
                        mSubImageCount++;
                    }
                }
                switch (mSubImageCount)
                {
                    case 2:
                        mSubHorizontalCount = 2;
                        mSubVerticalCount = 1;
                        break;
                    case 4:
                        mSubHorizontalCount = 2;
                        mSubVerticalCount = 2;
                        break;
                    case 6:
                        if ((mSubImageWidths[0] + mSubImageWidths[1]) == mImageWidth)
                        {
                            mSubHorizontalCount = 2;
                            mSubVerticalCount = 3;
                        }
                        else
                        {
                            mSubHorizontalCount = 3;
                            mSubVerticalCount = 2;
                        }
                        break;
                    default:
                        mSubHorizontalCount = mSubImageCount;
                        mSubVerticalCount = 1;
                        break;
                }
                if (mSubImageCount > 1)
                {
                    mTypeDescription = "Uncompressed Tiled Indexed Image";
                }
                else
                {
                    mTypeDescription = "Uncompressed Indexed Image";
                }

                ////original version

                //mImageDataOffset = 144;
                ////this type lies and claims to be 16-bit when really it's 32
                //mBPPImage = 32;
                ////manually set all parameters
                //mBPPAlpha = 8;
                //mBPPRed = 8;
                //mBPPGreen = 8;
                //mBPPBlue = 8;
                //mMaskAlpha = 0x000000FF;
                //mMaskRed = 0xFF000000;
                //mMaskGreen = 0x00FF0000;
                //mMaskBlue = 0x0000FF00;

                //mTypeDescription = "Uncompressed Tiled Indexed Image";
            }
            if ((mFileType == FileTypes.Defiance_PC))
            {
                mImageDataOffset = 128;
                mBPPAlpha = (int)tBPP1;
                mBPPRed = (int)tBPP4;
                mBPPGreen = (int)tBPP3;
                mBPPBlue = (int)tBPP2;
                //just to be difficult, this file type uses reversed masks
                mMaskAlpha = TransformMask(tMask1);
                mMaskRed = TransformMask(tMask4);
                mMaskGreen = TransformMask(tMask3);
                mMaskBlue = TransformMask(tMask2);
                mTypeDescription = "RLE-Compressed ARGB Image";
            }
            if ((mFileType == FileTypes.TRL_PSP_ARGB))
            {
                mImageDataOffset = 128;
                //these values are hardcoded because the header information is incorrect in this filetype
                //settings are based on the file type (1 or 2)
                //imageBitsPerPixel = (ushort)tBPP;
                mBPPImage = (ushort)16;

                if (tFileType == 1)
                {
                    mBPPAlpha = (ushort)0;
                    mBPPRed = (ushort)5;
                    mBPPGreen = (ushort)6;
                    mBPPBlue = (ushort)5;
                    mMaskAlpha = 0x00000000;
                    mMaskBlue = 0x0000F800;
                    mMaskGreen = 0x000007E0;
                    mMaskRed = 0x0000001F;
                    mTypeDescription = "Uncompressed Twiddled RGB 565 Image";
                }
                else
                {
                    mBPPAlpha = (ushort)4;
                    mBPPRed = (ushort)4;
                    mBPPGreen = (ushort)4;
                    mBPPBlue = (ushort)4;
                    mMaskAlpha = 0x0000F000;
                    mMaskBlue = 0x00000F00;
                    mMaskGreen = 0x000000F0;
                    mMaskRed = 0x0000000F;
                    mTypeDescription = "Uncompressed Twiddled ARGB 4444 Image";
                }
            }
            if ((mFileType == FileTypes.TRL_PSP_Indexed))
            {
                mImageDataOffset = 128;
                //these values are hardcoded because the header information is incorrect in this filetype
                //settings are based on the file type (1 or 2)
                //imageBitsPerPixel = (ushort)tBPP;
                mBPPImage = (ushort)32;

                mBPPAlpha = 8;
                mBPPRed = 8;
                mBPPGreen = 8;
                mBPPBlue = 8;
                mMaskAlpha = 0x000000FF;
                mMaskRed = 0xFF000000;
                mMaskGreen = 0x00FF0000;
                mMaskBlue = 0x0000FF00;
                mTypeDescription = "Uncompressed Twiddled Indexed Image";
            }   


            //transform the parameters now that we have all the data
            if (BPPImage == 32)
            {
                StartBitAlpha = (ushort)getBitPositionInDWord(mMaskAlpha);
                StartBitRed = (ushort)getBitPositionInDWord(mMaskRed);
                StartBitGreen = (ushort)getBitPositionInDWord(mMaskGreen);
                StartBitBlue = (ushort)getBitPositionInDWord(mMaskBlue);
            }
            else if (BPPImage == 16)
            {
                StartBitAlpha = (ushort)getBitPositionInWord(mMaskAlpha);
                StartBitRed = (ushort)getBitPositionInWord(mMaskRed);
                StartBitGreen = (ushort)getBitPositionInWord(mMaskGreen);
                StartBitBlue = (ushort)getBitPositionInWord(mMaskBlue);
            }

        }

        protected void GetSRTexParameters()
        {
            uint tWidth;
			uint tHeight;
            FileStream iStream;
            BinaryReader iReader;
            iStream = new FileStream(mFilePath, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);

			//get all the header parameters for an SRTex image
			iStream.Seek(28, SeekOrigin.Begin);
			tWidth = iReader.ReadUInt32() * 4;
			tHeight = iReader.ReadUInt32();

			iReader.Close();
			iStream.Close();
			mImageDataOffset = 36;
            mImageWidth = (ushort)tWidth;
            mImageHeight = (ushort)tHeight;
            mBPPImage = 4;
            mDataLength = (uint)(mImageWidth * mImageHeight / 2);
            mBPPAlpha = 0;
            mBPPRed = 4;
            mBPPGreen = 4;
            mBPPBlue = 4;
            mMaskAlpha = 0x00000000;
            mMaskRed = 0x00000004;
            mMaskGreen = 0x00000004;
            mMaskBlue = 0x00000004;
            mTypeDescription = "Soul Reaver Texture Array";
            mImageType = ImageTypes.Grey;
            mFileType = FileTypes.SRTex;
        }

        public void ReadImageFromFile()
        {
            switch (mFileType)
            {
                case FileTypes.SRTex:
                    ReadSoulReaverTextureArray();
                    break;
                case FileTypes.SR2_PS2_AirForge_ARGB:
                    ReadSingleARGBImageData();
                    break;
                case FileTypes.SR2_PS2_AirForge_Indexed:
                    ReadSR2PS2AirForgeIndexedImageData();
                    break;
                case FileTypes.SR2_PS2_ARGB:
                    ReadSR2PS2TiledRGBImageData();
                    break;
                case FileTypes.SR2_PS2_Indexed:
                    ReadSR2PS2TiledIndexedImageData();
                    break;
                case FileTypes.SR2_PC:
                    ReadSingleARGBImageData();
                    break;
                case FileTypes.Defiance_PS2_ARGB:
                    ReadDefiancePS2TiledImageData();
                    break;
                case FileTypes.Defiance_PS2_Indexed:
                    ReadDefiancePS2TiledIndexedImageData();
                    break;
                case FileTypes.Defiance_PC:
                    ReadDefiancePCImageData();
                    break;
                case FileTypes.TRL_PSP_ARGB:
                    ReadTwiddledSingleARGBImageData();
                    break;
                case FileTypes.TRL_PSP_Indexed:
                    ReadTRLPSPIndexedImageData();
                    break;
                default:
                    throw new Exception("Can't read an image in an unrecognized format.");
                    break;
            }
        }

        public void ReadSoulReaverTextureArray()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            byte[] rawData = ReadDataFromFile(mFilePath, mImageDataOffset, mDataLength);
            ImageData currentChannelData = new ImageData(this, mImageWidth, mImageHeight);
            currentChannelData.Get4BitChannelsFromRawBytes(rawData);

            mImage = currentChannelData.CreateImage();
        }

        public void ReadSingleARGBImageData()
        {
            byte[] rawData = ReadDataFromFile(mFilePath, mImageDataOffset, mImageWidth * mImageHeight * (BPPImage / 8));
            ImageData channelData = new ImageData(this);
            channelData.GetChannelsFromRawBytes(rawData);
            mImage = channelData.CreateImage();
        }

        public void ReadSR2PS2AirForgeIndexedImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            //get the palette first
            byte[] paletteData = ReadDataFromFile(mFilePath, 84, 512);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData);
            Bitmap tempPalette = paletteChannelData.CreateImage();

            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
            }

            byte[] rawData = ReadDataFromFile(mFilePath, 596, (mImageWidth * mImageHeight));
            //rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
            ImageData currentChannelData = new ImageData(this, mImageWidth, mImageHeight);
            currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0);
            //tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

            tempImage = currentChannelData.CreateImage();

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < mImageHeight; y++)
            {
                for (int x = 0; x < mImageWidth; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }

            mImage = finalImage;
        }

        public void ReadSR2PS2TiledIndexedImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            int subHorizontalCount = (ushort)(mImageWidth / 256);
            if ((subHorizontalCount * 256) < mImageWidth)
            {
                subHorizontalCount++;
            }
            //int subHorizontalCount = (ushort)(mImageWidth / 128);

            int[] subImageWidths = new int[32];
            int[] subImageHeights = new int[32];
            int imageCounter = 0;
            long positionCounter = 656;
            int vOffset = 0;
            ImageTransformer iTrans = new ImageTransformer();

            //get the palette first
            byte[] paletteData = ReadDataFromFile(mFilePath, 144, 512);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData);
            Bitmap tempPalette = paletteChannelData.CreateImage();
            
            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
            }

            //pretend we're dead^W a series of 8-bit images
            int rowCount = 0;
            do
            {
                int hOffset = 0;
                for (int j = 0; j < subHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, mImageWidth - hOffset);
                    long currentDataLength = ReadUInt32FromFile(mFilePath, positionCounter) * 16;

                    subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter]));

                    long subImageOffset = positionCounter + 16;

                    positionCounter += 16 + currentDataLength;

                    byte[] rawData = ReadDataFromFile(mFilePath, subImageOffset, currentDataLength);
                    rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                //vOffset += subImageHeights[imageCounter - 1];
                vOffset += subImageHeights[(rowCount * subHorizontalCount)];
                rowCount++;
            } while (vOffset < mImageHeight);

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < mImageHeight; y++)
            {
                for (int x = 0; x < mImageWidth; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }
 
            mImage = finalImage;
        }

        public void ReadSR2PS2TiledRGBImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            int subHorizontalCount = (ushort)(mImageWidth / 256);
            if ((subHorizontalCount * 256) < mImageWidth)
            {
                subHorizontalCount++;
            }
            //int subVerticalCount = (ushort)(mImageHeight / 256);
            //if ((subVerticalCount * 256) < mImageHeight)
            //{
            //    subVerticalCount++;
            //}
            //int subImageCount = subHorizontalCount * subVerticalCount;


            int[] subImageWidths = new int[32];
            int[] subImageHeights = new int[32];
            int imageCounter = 0;
            long positionCounter = 128;
            int vOffset = 0;

            int rowCount = 0;
            do
            {
                int hOffset = 0;
                for (int j = 0; j < subHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, mImageWidth - hOffset);
                    //get current data length
                    long currentDataLength = ReadUInt32FromFile(mFilePath, positionCounter) * 16;

                    subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter] * (mBPPImage / 8)));

                    long subImageOffset = positionCounter + 16;

                    positionCounter += 16 + currentDataLength;

                    byte[] rawData = ReadDataFromFile(mFilePath, subImageOffset, currentDataLength);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                //vOffset += subImageHeights[imageCounter - 1];
                vOffset += subImageHeights[(rowCount * subHorizontalCount)];
                rowCount++;
            } while (vOffset < mImageHeight);

            mImage = tempImage;
        }

        public void ReadTwiddledSingleARGBImageData()
        {
            byte[] rawData = ReadDataFromFile(mFilePath, mImageDataOffset, mImageWidth * mImageHeight * (BPPImage / 8));
            rawData = unSwizzleBytes(rawData, mImageWidth, mImageHeight, mBPPImage);
            ImageData channelData = new ImageData(this);
            channelData.GetChannelsFromRawBytes(rawData);
            mImage = channelData.CreateImage();
        }


        public void ReadDefiancePS2TiledImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            int[] subImageWidths = new int[mSubImageCount];
            int[] subImageHeights = new int[mSubImageCount];
            int imageCounter = 0;
            long positionCounter = 128;
            int vOffset = 0;

            for (int i = 0; i < mSubVerticalCount; i++)
            {
                int hOffset = 0;
                for (int j = 0; j < mSubHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, mImageWidth - hOffset);
                    subImageHeights[imageCounter] = (ushort)Math.Min(256, mImageHeight - vOffset);

                    long subImageOffset = positionCounter + 16;
                    long currentDataLength = (subImageWidths[imageCounter] * subImageHeights[imageCounter] * (BPPImage / 8));
                    positionCounter += 16 + currentDataLength;

                    byte[] rawData = ReadDataFromFile(mFilePath, subImageOffset, currentDataLength);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                vOffset += subImageHeights[imageCounter % mSubHorizontalCount]; 
            }
            mImage = tempImage;
        }

        public void ReadDefiancePS2TiledIndexedImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            //assuming images must be a multiple of 256 wide for now
            //int subHorizontalCount = (ushort)(mImageWidth / 256);
            //int subHorizontalCount = (ushort)(mImageWidth / 128);
            int subHorizontalCount = mSubHorizontalCount;

            int[] subImageWidths = new int[32];
            int[] subImageHeights = new int[32];
            int imageCounter = 0;
            long positionCounter = 1168;
            int vOffset = 0;

            //get the palette first
            byte[] paletteData = ReadDataFromFile(mFilePath, 144, 1024);
            //paletteData = FlipUInt32EndianOrder(paletteData);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData);
            Bitmap tempPalette = paletteChannelData.CreateImage();

            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
                //palette.SetPixel(i, 0, tempPalette.GetPixel(i, 0));
            }

            //pretend we're dead^W a series of 8-bit images
            //for (int i = 0; i < mSubVerticalCount; i++)
            int rowCount = 0;
            do
            {
                int hOffset = 0;
                for (int j = 0; j < mSubHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, mImageWidth - hOffset);
                    //subImageHeights[imageCounter] = (ushort)Math.Min(256, mImageHeight - vOffset);

                    long subImageOffset = positionCounter + 16;
                    //long currentDataLength = (subImageWidths[imageCounter] * subImageHeights[imageCounter]);
                    long currentDataLength = ReadUInt32FromFile(mFilePath, positionCounter) * 16;
                    positionCounter += 16 + currentDataLength;

                    subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter]));

                    byte[] rawData = ReadDataFromFile(mFilePath, subImageOffset, currentDataLength);
                    rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                vOffset += subImageHeights[rowCount * subHorizontalCount];
                rowCount++;
            } while (vOffset < mImageHeight) ;

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < mImageHeight; y++)
            {
                for (int x = 0; x < mImageWidth; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }

            mImage = finalImage;
        }

        //public void ReadDefiancePS2TiledIndexedImageData()
        //{
        //    Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

        //    //assuming images must be a multiple of 256 wide for now
        //    int subHorizontalCount = (ushort)(mImageWidth / 256);
        //    //int subHorizontalCount = (ushort)(mImageWidth / 128);

        //    int[] subImageWidths = new int[32];
        //    int[] subImageHeights = new int[32];
        //    int imageCounter = 0;
        //    long positionCounter = 1168;
        //    int vOffset = 0;
        //    ImageTransformer iTrans = new ImageTransformer();

        //    //get the palette first
        //    byte[] paletteData = ReadDataFromFile(mFilePath, 144, 1024);
        //    //paletteData = FlipUInt32EndianOrder(paletteData);
        //    ImageData paletteChannelData = new ImageData(this, 256, 1);
        //    paletteChannelData.GetChannelsFromRawBytes(paletteData);
        //    Bitmap tempPalette = paletteChannelData.CreateImage();

        //    //re-arrange the palette colours to match how the PS2 stores them
        //    Bitmap palette = new Bitmap(tempPalette);
        //    int[] clutIndex = BuildIDTex8ClutIndex();
        //    for (int i = 0; i < 256; i++)
        //    {
        //        palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
        //        //palette.SetPixel(i, 0, tempPalette.GetPixel(i, 0));
        //    }

        //    //pretend we're dead^W a series of 8-bit images
        //    do
        //    {
        //        int hOffset = 0;
        //        for (int j = 0; j < subHorizontalCount; j++)
        //        {
        //            subImageWidths[imageCounter] = (ushort)Math.Min(256, mImageWidth - hOffset);
        //            //get current data length
        //            long currentDataLength = ReadUInt32FromFile(mFilePath, positionCounter) * 16;

        //            subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter]));

        //            long subImageOffset = positionCounter + 16;

        //            positionCounter += 16 + currentDataLength;

        //            byte[] rawData = ReadDataFromFile(mFilePath, subImageOffset, currentDataLength);
        //            rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
        //            ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
        //            currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0);
        //            tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

        //            hOffset += subImageWidths[imageCounter];
        //            imageCounter++;
        //        }
        //        vOffset += subImageHeights[imageCounter - 1];
        //    } while (vOffset < mImageHeight);

        //    //use the resulting image as an index into the palette
        //    Bitmap finalImage = new Bitmap(tempImage);
        //    for (int y = 0; y < mImageHeight; y++)
        //    {
        //        for (int x = 0; x < mImageWidth; x++)
        //        {
        //            //all of the channels in the tempImage except Alpha will have the index #. I am picking
        //            //green arbitrarily.
        //            finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
        //        }
        //    }

        //    mImage = finalImage;
        //}

        public void ReadDefiancePCImageData()
        {
            byte[] rawData = ReadDefiancePCDataFromFile(mFilePath, mImageDataOffset, mImageWidth * mImageHeight * (BPPImage / 8));
            ImageData channelData = new ImageData(this);
            channelData.GetChannelsFromRawBytes(rawData);
            mImage = channelData.CreateImage();
        }

        public void ReadTRLPSPIndexedImageData()
        {
            Bitmap tempImage = new Bitmap(ImageWidth, ImageHeight);

            //get the palette first
            byte[] paletteData = ReadDataFromFile(mFilePath, 128, 1024);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData);
            Bitmap tempPalette = paletteChannelData.CreateImage();

            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(i, 0));
                //palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
            }

            byte[] rawData = ReadDataFromFile(mFilePath, 1152, (mImageWidth * mImageHeight));
            rawData = unSwizzleBytes(rawData, mImageWidth, mImageHeight, 8);
            //rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
            ImageData currentChannelData = new ImageData(this, mImageWidth, mImageHeight);
            currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0);
            //tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

            tempImage = currentChannelData.CreateImage();

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < mImageHeight; y++)
            {
                for (int x = 0; x < mImageWidth; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }

            mImage = finalImage;
        }

        protected Bitmap compositeImages(Bitmap mainImage, Bitmap addImage, int x, int y)
        {
            Bitmap tcImage;
            Color tColour;
            int tA, tR, tG, tB;
            int xRep, yRep;

            xRep = Math.Min(addImage.Width, mainImage.Width - x);
            yRep = Math.Min(addImage.Height, mainImage.Height - y);

            tcImage = new Bitmap(mainImage);

            for (int nY = 0; nY < yRep; nY++)
            {
                for (int nX = 0; nX < xRep; nX++)
                {
                    tColour = addImage.GetPixel(nX, nY);
                    tA = tColour.A;
                    tR = tColour.R;
                    tG = tColour.G;
                    tB = tColour.B;
                    tcImage.SetPixel(nX + x, nY + y, Color.FromArgb(tA, tR, tG, tB));
                }
            }

            return tcImage;
        }

        protected byte[] unSwizzleBytes(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int colCount = imgWidth / 16 * (imgBPP / 8);
            int rowCount = imgHeight / 8;

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int i = 0; i < rowCount; i++)
            {
                for (int j = 0; j < colCount; j++)
                {
                    for (int k = 0; k < 128; k++)
                    {
                        currentOutPos = (i * colCount * 128) + (j * 16) + (k % 16) + ((k / 16) * colCount * 16);
                        outData[currentOutPos] = inData[currentInPos];
                        currentInPos++;
                    }
                }
            }

            return outData;
        }

        protected byte[] unSwizzleBytes(byte[] inData, int imgWidth, int imgHeight, int imgBPP, int colX, int colY)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int colCount = imgWidth / colX * (imgBPP / 8);
            int rowCount = imgHeight / colY;
            int colProduct = colX * colY;

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int i = 0; i < rowCount; i++)
            {
                for (int j = 0; j < colCount; j++)
                {
                    for (int k = 0; k < colProduct; k++)
                    {
                        currentOutPos = (i * colCount * colProduct) + (j * colX) + (k % colX) + ((k / colX) * colCount * colX);
                        outData[currentOutPos] = inData[currentInPos];
                        currentInPos++;
                    }
                }
            }

            return outData;
        }

        protected byte[] FlipUInt16EndianOrder(byte[] inData)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            for (int i = 0; i < inData.GetUpperBound(0); i += 2)
            {
                outData[i] = inData[i + 1];
                outData[i + 1] = inData[i];
            }

            return outData;
        }

        protected byte[] FlipUInt32EndianOrder(byte[] inData)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            for (int i = 0; i < inData.GetUpperBound(0); i += 4)
            {
                outData[i] = inData[i + 3];
                outData[i + 1] = inData[i + 2];
                outData[i + 2] = inData[i + 1];
                outData[i + 3] = inData[i];
            }

            return outData;
        }

        protected byte[] unPS2SR2Swizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            //return inData;
            byte[] outData;
            outData = new byte[inData.Length];

            //int z = 0;
            //for (int y = 0; y < imgHeight; y++)
            //{
            //    for (int x = 1; x < imgWidth; x += 2)
            //    {
            //        outData[(y * imgWidth) + x] = inData[z];
            //        z++;
            //    }
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        outData[(y * imgWidth) + x] = inData[z];
            //        z++;
            //    }
            //}


            //this one is really close, but seems to need each meta-block of 4x4 pixels rotated too
            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        int inPos = (imgWidth * y) + x;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * y) + x + 1;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * (y + 1)) + x;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * (y + 1)) + x + 1;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //}

            int[] pixelIndex02 = new int[]
            {
                0, 1, 4, 5, 8, 9, 12, 13,       0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15,     2, 3, 6, 7, 10, 11, 14, 15,

                8, 9, 12, 13, 0, 1, 4, 5,       8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7,     10, 11, 14, 15, 2, 3, 6, 7 
            };

            int[] pixelIndex13 = new int[]
            {
                8, 9, 12, 13, 0, 1, 4, 5,       8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7,     10, 11, 14, 15, 2, 3, 6, 7, 

                0, 1, 4, 5, 8, 9, 12, 13,       0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15,     2, 3, 6, 7, 10, 11, 14, 15
            };

            //int[] subIndex = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,          32, 32, 32, 32, 32, 32, 32, 32,
            //    0, 0, 0, 0, 0, 0, 0, 0,          32, 32, 32, 32, 32, 32, 32, 32,

            //    16, 16, 16, 16, 16, 16, 16, 16,      48, 48, 48, 48, 48, 48, 48, 48,
            //    16, 16, 16, 16, 16, 16, 16, 16,      48, 48, 48, 48, 48, 48, 48, 48
            //};

            int[] subIndex = new int[]
            {
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,

                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8
            };

            //int[] xIndex = new int[]
            //{
            //    0, 1, 
            //    0, 1
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0,
            //    1, 1
            //};

            //int[] xIndex = new int[]
            //{
            //    1, 0, 
            //    1, 0
            //};

            //int[] yIndex = new int[]
            //{
            //    1, 1,
            //    0, 0
            //};

            //int[] xIndex = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,

            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11,
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,              0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1,              1, 1, 1, 1, 1, 1, 1, 1, 
            //    2, 2, 2, 2, 2, 2, 2, 2,              2, 2, 2, 2, 2, 2, 2, 2, 
            //    3, 3, 3, 3, 3, 3, 3, 3,              3, 3, 3, 3, 3, 3, 3, 3
            //};

            //int[] xIndexA = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15
            //};

            //int[] xIndexA = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,
            //    0, 1, 2, 3, 4, 5, 6, 7
            //};

            //int[] xIndexB = new int[]
            //{
            //    4, 5, 6, 7, 0, 1, 2, 3,
            //    4, 5, 6, 7, 0, 1, 2, 3
            //};

            //int[] yIndex2 = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1
            //};

            //int[] xIndexB = new int[]
            //{
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11,
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11
            //};

            //int[] yIndex2 = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,              0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1,              1, 1, 1, 1, 1, 1, 1, 1, 
            //};

            //int[] blockIndex = new int[]
            //{
            //    0, 1, 4, 5,         16, 17, 20, 21,
            //    2, 3, 6, 7,         18, 19, 22, 23,
            //    8, 9, 12, 13,       24, 25, 28, 29,
            //    10, 11, 14, 15,     26, 27, 30, 31
            //};

            //int[] blockIndex16 = new int[]
            //{
            //    0, 1, 4, 5,
            //    2, 3, 6, 7,
            //    8, 9, 12, 13,
            //    10, 11, 14, 15
            //};

            //int[] blockIndex4 = new int[]
            //{
            //    0, 1,
            //    2, 3
            //};

            //int[] xIndex = new int[]
            //{
            //    0, 1, 2, 3, 
            //    0, 1, 2, 3,
            //    0, 1, 2, 3,
            //    0, 1, 2, 3
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0, 0, 0,
            //    1, 1, 1, 1,
            //    2, 2, 2, 2,
            //    3, 3, 3, 3
            //};

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        for (int i = 0; i < 4; i++)
            //        {
            //            int inPos = (imgWidth * (y + yIndex[i])) + x + xIndex[i];
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //int[] bitMap = new int[]
            //{
            //    6, 7, 8, 9,
            //    5, 0, 1, 10,
            //    4, 3, 2, 11,
            //    15, 14, 13, 12
            //};

            int[] bitMap0 = new int[]
            {
                0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15
            };

            int[] bitMap1 = new int[]
            {
                8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7
            };

            //int inPos = 0;
            int globalX = 0;
            int globalY = 0;
            int chunkX = 8;
            int chunkY = 2;
            int chunkSize = chunkX * chunkY;
            int[] bitMap = bitMap0;
            int whichBitMap = 0;
            for (int inPos = 0; inPos < inData.Length; inPos += chunkSize)
            {
                //byte[] localBytes = new byte[chunkSize];
                //for (int i = 0; i < localBytes.Length; i++)
                //{
                //    localBytes[i] = inData[inPos + i];
                //}
                int iterator = 0;
                for (int y = 0; y < chunkY; y++)
                {
                    for (int x = 0; x < chunkX; x++)
                    {
                        int mappedPosition = bitMap[(y * chunkX) + x];
                        int localX = globalX + (mappedPosition % chunkX);
                        int localY = globalY + (mappedPosition / chunkX);
                        int outPos = (localY * imgWidth) + localX;
                        outData[outPos] = inData[inPos + iterator];
                        iterator++;
                    }
                }
                globalX += chunkX;
                if (globalX >= imgWidth)
                {
                    whichBitMap++;
                    if (whichBitMap > 3)
                    {
                        whichBitMap = 0;
                    }
                    switch (whichBitMap)
                    {
                        case 0:
                            bitMap = bitMap1;
                            break;
                        case 1:
                            bitMap = bitMap1;
                            break;
                        case 2:
                            bitMap = bitMap1;
                            break;
                        case 3:
                            bitMap = bitMap1;
                            break;
                    }
                    globalX = 0;
                    globalY += chunkY;
                }
            }
            //int inPosBase = 0;
            //for (int y = 0; y < imgHeight; y += 8)
            //{
            //    inPosBase = y * imgWidth;
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int j = 0; j < 64; j++)
            //        {
            //            //int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            int inPos = inPosBase + x + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            outData[outPos] = inData[inPos];
            //            //outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //        //inPosBase += 64;
            //    }
            //    inPosBase = (y + 4) * imgWidth;
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int j = 0; j < 64; j++)
            //        {
            //            //int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            int inPos = inPosBase + x + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            outData[outPos] = inData[inPos];
            //            //outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //        //inPosBase += 64;
            //    } 
            //    //for (int i = 0; i < inData.Length; i += 128)
            //    //{
            //    //    inPosBase = i;
            //    //    for (int j = 0; j < 64; j++)
            //    //    {
            //    //        int inPos = inPosBase + pixelIndex02[j] + subIndex[j];
            //    //        outData[outPos] = inData[inPos];
            //    //        //outData[inPos] = inData[outPos];
            //    //        outPos++;
            //    //    }
            //    //    inPosBase += 64;
            //    //    for (int j = 0; j < 64; j++)
            //    //    {
            //    //        int inPos = inPosBase + pixelIndex13[j] + subIndex[j];
            //    //        outData[outPos] = inData[inPos];
            //    //        //outData[inPos] = inData[outPos];
            //    //        outPos++;
            //    //    }
            //    //}
            //}
            //for (int i = 0; i < inData.Length; i += 128)
            //{
            //    inPosBase = i;
            //    for (int j = 0; j < 64; j++)
            //    {
            //        int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //    inPosBase += 64;
            //    for (int j = 0; j < 64; j++)
            //    {
            //        int inPos = inPosBase + pixelIndex13[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //}
            //int inPos = 0;
            //for (int y = 0; y < imgHeight; y += 4)
            //{
            //    for (int i = 0; i < (imgWidth * 2); i += 16)
            //    {
            //        int startX = i / 2;
            //        for (int k = 0; k < 4; k++)
            //        {
            //            int currentX = startX + (k * 2);
            //            for (int j = 0; j < 4; j++)
            //            {
            //                int outPos = (y * imgWidth) + (yIndex[j] * imgWidth) + currentX + xIndex[j];
            //                outData[outPos] = inData[inPos];
            //                inPos++;
            //            }
            //        }
            //    }
            //    for (int i = 0; i < (imgWidth * 2); i += 16)
            //    {
            //        int startX = i / 2;
            //        for (int k = 0; k < 4; k++)
            //        {
            //            int currentX = startX + (k * 2);
            //            for (int j = 0; j < 4; j++)
            //            {
            //                int outPos = ((y + 2) * imgWidth) + (yIndex[j] * imgWidth) + currentX + xIndex[j];
            //                outData[outPos] = inData[inPos];
            //                inPos++;
            //            }
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 4)
            //    {
            //        for (int j = 0; j < 4; j++)
            //        {
            //            for (int i = 0; i < 4; i++)
            //            {
            //                int inPos = (imgWidth * (y + yIndex[i])) + x + xIndex[i];
            //                inPos += (xIndex[j] * 2);
            //                //inPos += (yIndex[j] * 2);
            //                outData[inPos] = inData[outPos];
            //                outPos++;
            //            }
            //        }
            //    }
            //}

            //int numBlocks = (inData.Length / (128 * 64));
            //for (int blockNum = 0; blockNum < numBlocks; blockNum++)
            //{
            //    for (int colNum = 0; colNum < 8; colNum++)
            //    {
            //        for (int z = 0; z < (2 * 16); z++)
            //        {
            //            int xFactor = 0;
            //            if (colNum % 2 == 1)
            //            {
            //                xFactor = xIndexB[z];
            //            }
            //            else
            //            {
            //                xFactor = xIndexA[z];
            //            }
            //            xFactor += (yIndex[z] * imgWidth);
            //            int inPos = 0;
            //            inPos = (blockNum * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            if (numBlocks < 5)
            //            {
            //                inPos = (blockIndex4[blockNum] * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            }
            //            else
            //            {
            //                inPos = (blockIndex16[blockNum] * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            }
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //int arraySwitch = 0;

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int i = 0; i < (2 * 16); i++)
            //        {
            //            int xFactor = 0;
            //            switch (arraySwitch)
            //            {
            //                case 0:
            //                    xFactor = xIndexA[i];
            //                    break;
            //                case 1:
            //                    xFactor = xIndexB[i];
            //                    break;
            //                case 2:
            //                    xFactor = xIndexB[i];
            //                    break;
            //                case 3:
            //                    xFactor = xIndexA[i];
            //                    break;
            //                default:
            //                    throw new Exception("Error");
            //                    break;
            //            }
            //            arraySwitch++;
            //            if (arraySwitch > 3)
            //            {
            //                arraySwitch = 0;
            //            }
            //            int inPos = (imgWidth * (y + yIndex[i])) + x + xFactor;
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight / 4; y += 4)
            //{
            //    for (int x = 0; x < imgWidth / 16; x += 16)
            //    {
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = pixelIndex02[i] + subIndex[i];
            //            inPos += (y * imgWidth) + x;
            //            outData[inPos] = inData[outPos];
            //        }
            //    }
            //}

            //int byteNum = 0;

            //for (int blockY = 0; blockY < imgHeight / 16; blockY++)
            //{
            //    for (int blockX = 0; blockX < imgWidth / 16; blockX++)
            //    {
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex02[i] + subIndex[i];
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex13[i] + subIndex[i] + 64;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex02[i] + subIndex[i] + 128;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex13[i] + subIndex[i] + 192;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //    }
            //}

            //for (int block = 0; block < inData.Length / 256; block++)
            //{
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex02[i] + subIndex[i]];
            //        outData[(block * 256) + pixelIndex02[i] + subIndex[i]] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex13[i] + subIndex[i] + 64];
            //        outData[(block * 256) + pixelIndex13[i] + subIndex[i] + 64] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex02[i] + subIndex[i] + 128];
            //        outData[(block * 256) + pixelIndex02[i] + subIndex[i] + 128] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex13[i] + subIndex[i] + 192];
            //        outData[(block * 256) + pixelIndex13[i] + subIndex[i] + 192] = inData[byteNum];
            //        byteNum++;
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 32;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = div; xx < imgWidth; xx += numDivs)
            //        {
            //            outData[(y * imgWidth) + x] = inData[(y * imgWidth + xx)];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 4;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = div; xx < imgWidth; xx += numDivs)
            //        {
            //            outData[(y * imgWidth) + xx] = inData[(y * imgWidth + x)];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 64;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = 0; xx < divWidth; xx++)
            //        {
            //            int outPixelX = (div * divWidth) + (divWidth - 1 - xx);
            //            int inPixelX = xx + (div * divWidth);
            //            //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 4;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = 0; xx < divWidth; xx++)
            //        {
            //            int outPixelX = (div * divWidth) + (divWidth - 1 - xx);
            //            int inPixelX = xx + (div * divWidth);
            //            //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //            x++;
            //        }
            //    }
            //}

            //int outPos = 0;

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    int divCount = 1;
            //    int divWidth = imgWidth / divCount;
            //    for (int div = 0; div < divCount; div++)
            //    {
            //        //for (int x = 0; x < (divWidth / 2); x++)
            //        //{
            //        //    outData[outPos] = inData[(y * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}
            //        //for (int x = 0; x < (divWidth / 2); x++)
            //        //{
            //        //    outData[outPos] = inData[((y + 1) * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}

            //        //for (int x = (divWidth / 2); x < divWidth; x++)
            //        //{
            //        //    outData[outPos] = inData[(y * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}
            //        //for (int x = (divWidth / 2); x < divWidth; x++)
            //        //{
            //        //    outData[outPos] = inData[((y + 1) * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}

            //        for (int x = 0; x < (divWidth / 2); x++)
            //        {
            //            outData[(y * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }
            //        for (int x = 0; x < (divWidth / 2); x++)
            //        {
            //            outData[((y + 1) * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }

            //        for (int x = (divWidth / 2); x < divWidth; x++)
            //        {
            //            outData[(y * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }
            //        for (int x = (divWidth / 2); x < divWidth; x++)
            //        {
            //            outData[((y + 1) * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }




            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    //int x = 0;
            //    int numDivs = 8;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        int numSubDivs = 8;
            //        int subDivWidth = divWidth / numSubDivs;
            //        for (int subDiv = 0; subDiv < numSubDivs; subDiv++)
            //        {
            //            int x = 0;
            //            for (int xx = 0; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //                x++;
            //            }
            //            for (int xx = 1; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //                x++;
            //            }

            //            //for (int xx = 0; xx < subDivWidth; xx++)
            //            //{
            //            //    int outPixelX = subDivWidth - 1 - xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            //    outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            //    x++;
            //            //}
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    //int x = 0;
            //    int numDivs = 16;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        int numSubDivs = 2;
            //        int subDivWidth = divWidth / numSubDivs;
            //        for (int subDiv = 0; subDiv < numSubDivs; subDiv++)
            //        {
            //            int x = 0;
            //            for (int xx = 0; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //                x++;
            //            }
            //            for (int xx = 1; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //                x++;
            //            }

            //            //for (int xx = 0; xx < subDivWidth; xx++)
            //            //{
            //            //    int outPixelX = subDivWidth - 1 - xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            //    outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            //    x++;
            //            //}
            //        }
            //    }
            //}

            //int currentInPos = 0;
            //int currentOutPos = 0;

            //int colCount = imgWidth / 128;
            //int rowCount = imgHeight / 64;

            //for (int i = 0; i < rowCount; i++)
            //{
            //    for (int j = 0; j < colCount; j++)
            //    {
            //        for (int k = 0; k < 128; k++)
            //        {
            //            currentOutPos = (i * colCount * 128) + (j * 16) + (k % 16) + ((k / 16) * colCount * 16);
            //            outData[currentOutPos] = inData[currentInPos];
            //            currentInPos++;
            //        }
            //    }
            //}

            //// 2/2 is close
            //int xFactor = 16;
            //int yFactor = 16;
            //int blockSize = xFactor * yFactor;

            //int colCount = imgWidth / xFactor * (imgBPP / 8);
            //int rowCount = imgHeight / yFactor;

            //int[][] pixelIndex = BuildPSMT8PixelIndex();


            //for (int i = 0; i < rowCount; i++)
            //{
            //    for (int j = 0; j < colCount; j++)
            //    {
            //        for (int k = 0; k < blockSize; k++)
            //        {
            //            //currentOutPos = (i * colCount * blockSize) + (j * xFactor) + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            //currentInPos = (i * colCount * blockSize) + (j * xFactor) + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            //              ^^^for each entire row        ^h to block     ^x in block      ^y in block    
            //            //currentOutPos = (i * colCount * blockSize) + (j * xFactor) + (k % 64) + ((k / 64) * colCount * 16);
            //            int currentBasePos = (i * colCount * blockSize) + (j * xFactor);
            //            currentOutPos = currentBasePos + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            currentInPos = currentBasePos + (pixelIndex[k % 4][k % 64]) + ((pixelIndex[k / xFactor][k] / xFactor) * colCount * xFactor);//pixelIndex[k / 16][k];
            //            outData[currentOutPos] = inData[currentInPos];
            //            currentOutPos++;
            //        }
            //    }
            //}

            return outData;
        }

        protected byte[] unPS2DefianceSwizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            //return inData;
            byte[] outData;
            outData = new byte[inData.Length];

            Hashtable map1 = new Hashtable();
            Hashtable map2 = new Hashtable();

            map1.Add(0, new Point(0, 0));
            map1.Add(1, new Point(4, 1));
            map1.Add(2, new Point(8, 0));
            map1.Add(3, new Point(12, 1));
            map1.Add(4, new Point(1, 0));
            map1.Add(5, new Point(5, 1));
            map1.Add(6, new Point(9, 0));
            map1.Add(7, new Point(13, 1));
            map1.Add(8, new Point(2, 0));
            map1.Add(9, new Point(6, 1));
            map1.Add(10, new Point(10, 0));
            map1.Add(11, new Point(14, 1));
            map1.Add(12, new Point(3, 0));
            map1.Add(13, new Point(7, 1));
            map1.Add(14, new Point(11, 0));
            map1.Add(15, new Point(15, 1));
            map1.Add(16, new Point(4, 0));
            map1.Add(17, new Point(0, 1));
            map1.Add(18, new Point(12, 0));
            map1.Add(19, new Point(8, 1));
            map1.Add(20, new Point(5, 0));
            map1.Add(21, new Point(1, 1));
            map1.Add(22, new Point(13, 0));
            map1.Add(23, new Point(9, 1));
            map1.Add(24, new Point(6, 0));
            map1.Add(25, new Point(2, 1));
            map1.Add(26, new Point(14, 0));
            map1.Add(27, new Point(10, 1));
            map1.Add(28, new Point(7, 0));
            map1.Add(29, new Point(3, 1));
            map1.Add(30, new Point(15, 0));
            map1.Add(31, new Point(11, 1));

            map2.Add(0, new Point(4, 0));
            map2.Add(1, new Point(0, 1));
            map2.Add(2, new Point(12, 0));
            map2.Add(3, new Point(8, 1));
            map2.Add(4, new Point(5, 0));
            map2.Add(5, new Point(1, 1));
            map2.Add(6, new Point(13, 0));
            map2.Add(7, new Point(9, 1));
            map2.Add(8, new Point(6, 0));
            map2.Add(9, new Point(2, 1));
            map2.Add(10, new Point(14, 0));
            map2.Add(11, new Point(10, 1));
            map2.Add(12, new Point(7, 0));
            map2.Add(13, new Point(3, 1));
            map2.Add(14, new Point(15, 0));
            map2.Add(15, new Point(11, 1));
            map2.Add(16, new Point(0, 0));
            map2.Add(17, new Point(4, 1));
            map2.Add(18, new Point(8, 0));
            map2.Add(19, new Point(12, 1));
            map2.Add(20, new Point(1, 0));
            map2.Add(21, new Point(5, 1));
            map2.Add(22, new Point(9, 0));
            map2.Add(23, new Point(13, 1));
            map2.Add(24, new Point(2, 0));
            map2.Add(25, new Point(6, 1));
            map2.Add(26, new Point(10, 0));
            map2.Add(27, new Point(14, 1));
            map2.Add(28, new Point(3, 0));
            map2.Add(29, new Point(7, 1));
            map2.Add(30, new Point(11, 0));
            map2.Add(31, new Point(15, 1));

            int[] map = new int[]
            {        
                           
                0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15,
                16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31
            };

            int inPos = 0;

            for (int y = 0; y < imgHeight; y += 8)
            {
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y;
                        int xPos = x;

                        Point mapPoint = (Point)map1[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 2;
                        int xPos = x;

                        Point mapPoint = (Point)map1[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 4;
                        int xPos = x;

                        Point mapPoint = (Point)map2[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 6;
                        int xPos = x;

                        Point mapPoint = (Point)map2[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        inPos++;
                    }
                }
            }

            //flip rows
            for (int y = 0; y < imgHeight; y += 4)
            {
                byte[] rowBuffer = new byte[imgWidth];
                //copy second row to buffer
                for (int i = 0; i < imgWidth; i++)
                {
                    rowBuffer[i] = outData[((y + 1) * imgWidth) + i];
                }
                //copy third row to second
                for (int i = 0; i < imgWidth; i++)
                {
                    outData[((y + 1) * imgWidth) + i] = outData[((y + 2) * imgWidth) + i];
                }
                //copy buffer to third row
                for (int i = 0; i < imgWidth; i++)
                {
                    outData[((y + 2) * imgWidth) + i] = rowBuffer[i];
                }
            }

            return outData;
        }

        //protected byte[] unPS2DefianceSwizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        //{
        //    //return inData;
        //    byte[] outData;
        //    outData = new byte[inData.Length];

        //    //int[] masterMapX1 = new int[]
        //    //{
        //    //    0,  0,  1,  1,  2,  2,  3,  3,     4,   4,  5,  5,  6,  6,  7,  7,     

        //    //    8,   8,  9,  9, 10, 10, 11, 11,     12, 12, 13, 13, 14, 14, 15, 15
        //    //};

        //    //int[] masterMapY1 = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    //};


        //    //int[] masterMapX2 = new int[]
        //    //{
        //    //    0,  0,  1,  1,  2,  2,  3,  3,     4,   4,  5,  5,  6,  6,  7,  7,     

        //    //    8,   8,  9,  9, 10, 10, 11, 11,     12, 12, 13, 13, 14, 14, 15, 15
        //    //};

        //    //int[] masterMapY2 = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    //};

        //    //Hashtable map1 = new Hashtable();
        //    //Hashtable map2 = new Hashtable();

        //    //map1.Add(0, new Point(0, 0));
        //    //map1.Add(1, new Point(0, 1));
        //    //map1.Add(2, new Point(1, 0));
        //    //map1.Add(3, new Point(1, 1));
        //    //map1.Add(4, new Point(2, 0));
        //    //map1.Add(5, new Point(2, 1));
        //    //map1.Add(6, new Point(3, 0));
        //    //map1.Add(7, new Point(3, 1));
        //    //map1.Add(8, new Point(4, 0));
        //    //map1.Add(9, new Point(4, 1));
        //    //map1.Add(10, new Point(5, 0));
        //    //map1.Add(11, new Point(5, 1));
        //    //map1.Add(12, new Point(6, 0));
        //    //map1.Add(13, new Point(6, 1));
        //    //map1.Add(14, new Point(7, 0));
        //    //map1.Add(15, new Point(7, 1));
        //    //map1.Add(16, new Point(8, 0));
        //    //map1.Add(17, new Point(8, 1));
        //    //map1.Add(18, new Point(9, 0));
        //    //map1.Add(19, new Point(9, 1));
        //    //map1.Add(20, new Point(10, 0));
        //    //map1.Add(21, new Point(10, 1));
        //    //map1.Add(22, new Point(11, 0));
        //    //map1.Add(23, new Point(11, 1));
        //    //map1.Add(24, new Point(12, 0));
        //    //map1.Add(25, new Point(12, 1));
        //    //map1.Add(26, new Point(13, 0));
        //    //map1.Add(27, new Point(13, 1));
        //    //map1.Add(28, new Point(14, 0));
        //    //map1.Add(29, new Point(14, 1));
        //    //map1.Add(30, new Point(15, 0));
        //    //map1.Add(31, new Point(15, 1));

        //    //map2.Add(0, new Point(0, 0));
        //    //map2.Add(1, new Point(0, 1));
        //    //map2.Add(2, new Point(1, 0));
        //    //map2.Add(3, new Point(1, 1));
        //    //map2.Add(4, new Point(2, 0));
        //    //map2.Add(5, new Point(2, 1));
        //    //map2.Add(6, new Point(3, 0));
        //    //map2.Add(7, new Point(3, 1));
        //    //map2.Add(8, new Point(4, 0));
        //    //map2.Add(9, new Point(4, 1));
        //    //map2.Add(10, new Point(5, 0));
        //    //map2.Add(11, new Point(5, 1));
        //    //map2.Add(12, new Point(6, 0));
        //    //map2.Add(13, new Point(6, 1));
        //    //map2.Add(14, new Point(7, 0));
        //    //map2.Add(15, new Point(7, 1));
        //    //map2.Add(16, new Point(8, 0));
        //    //map2.Add(17, new Point(8, 1));
        //    //map2.Add(18, new Point(9, 0));
        //    //map2.Add(19, new Point(9, 1));
        //    //map2.Add(20, new Point(10, 0));
        //    //map2.Add(21, new Point(10, 1));
        //    //map2.Add(22, new Point(11, 0));
        //    //map2.Add(23, new Point(11, 1));
        //    //map2.Add(24, new Point(12, 0));
        //    //map2.Add(25, new Point(12, 1));
        //    //map2.Add(26, new Point(13, 0));
        //    //map2.Add(27, new Point(13, 1));
        //    //map2.Add(28, new Point(14, 0));
        //    //map2.Add(29, new Point(14, 1));
        //    //map2.Add(30, new Point(15, 0));
        //    //map2.Add(31, new Point(15, 1));

        //    Hashtable map1 = new Hashtable();
        //    Hashtable map2 = new Hashtable();

        //    map1.Add(0, new Point(0, 0));
        //    map1.Add(1, new Point(0, 1));
        //    map1.Add(2, new Point(1, 0));
        //    map1.Add(3, new Point(1, 1));
        //    map1.Add(4, new Point(2, 0));
        //    map1.Add(5, new Point(2, 1));
        //    map1.Add(6, new Point(3, 0));
        //    map1.Add(7, new Point(3, 1));
        //    map1.Add(8, new Point(4, 0));
        //    map1.Add(9, new Point(4, 1));
        //    map1.Add(10, new Point(5, 0));
        //    map1.Add(11, new Point(5, 1));
        //    map1.Add(12, new Point(6, 0));
        //    map1.Add(13, new Point(6, 1));
        //    map1.Add(14, new Point(7, 0));
        //    map1.Add(15, new Point(7, 1));
        //    map1.Add(16, new Point(0, 2));
        //    map1.Add(17, new Point(0, 3));
        //    map1.Add(18, new Point(1, 2)); //
        //    map1.Add(19, new Point(1, 3));
        //    map1.Add(20, new Point(2, 2));
        //    map1.Add(21, new Point(2, 3));
        //    map1.Add(22, new Point(3, 2));
        //    map1.Add(23, new Point(3, 3));
        //    map1.Add(24, new Point(4, 2));
        //    map1.Add(25, new Point(4, 3));
        //    map1.Add(26, new Point(5, 2));
        //    map1.Add(27, new Point(5, 3));
        //    map1.Add(28, new Point(6, 2));
        //    map1.Add(29, new Point(6, 3));
        //    map1.Add(30, new Point(7, 2));
        //    map1.Add(31, new Point(7, 3));
        //    map1.Add(32, new Point(8, 0));
        //    map1.Add(33, new Point(8, 1));
        //    map1.Add(34, new Point(9, 0));
        //    map1.Add(35, new Point(9, 1));
        //    map1.Add(36, new Point(10, 0));
        //    map1.Add(37, new Point(10, 1));
        //    map1.Add(38, new Point(11, 0));
        //    map1.Add(39, new Point(11, 1));
        //    map1.Add(40, new Point(12, 0));
        //    map1.Add(41, new Point(12, 1));
        //    map1.Add(42, new Point(13, 0));
        //    map1.Add(43, new Point(13, 1));
        //    map1.Add(44, new Point(14, 0));
        //    map1.Add(45, new Point(14, 1));
        //    map1.Add(46, new Point(15, 0));
        //    map1.Add(47, new Point(15, 1));
        //    map1.Add(48, new Point(8, 2));
        //    map1.Add(49, new Point(8, 3));
        //    map1.Add(50, new Point(9, 2));
        //    map1.Add(51, new Point(9, 3));
        //    map1.Add(52, new Point(10, 2)); //
        //    map1.Add(53, new Point(10, 3));
        //    map1.Add(54, new Point(11, 2));
        //    map1.Add(55, new Point(11, 3));
        //    map1.Add(56, new Point(12, 2));
        //    map1.Add(57, new Point(12, 3));
        //    map1.Add(58, new Point(13, 2));
        //    map1.Add(59, new Point(13, 3));
        //    map1.Add(60, new Point(14, 2));
        //    map1.Add(61, new Point(14, 3));
        //    map1.Add(62, new Point(15, 2));
        //    map1.Add(63, new Point(15, 3));

        //    map2.Add(0, new Point(0, 2));
        //    map2.Add(1, new Point(0, 3));
        //    map2.Add(2, new Point(1, 2));
        //    map2.Add(3, new Point(1, 3));
        //    map2.Add(4, new Point(2, 2));
        //    map2.Add(5, new Point(2, 3));
        //    map2.Add(6, new Point(3, 2));
        //    map2.Add(7, new Point(3, 3));
        //    map2.Add(8, new Point(4, 2));
        //    map2.Add(9, new Point(4, 3));
        //    map2.Add(10, new Point(5, 2));
        //    map2.Add(11, new Point(5, 3));
        //    map2.Add(12, new Point(6, 2));
        //    map2.Add(13, new Point(6, 3));
        //    map2.Add(14, new Point(7, 2));
        //    map2.Add(15, new Point(7, 3));
        //    map2.Add(16, new Point(0, 0));
        //    map2.Add(17, new Point(0, 1));
        //    map2.Add(18, new Point(1, 0));
        //    map2.Add(19, new Point(1, 1));
        //    map2.Add(20, new Point(2, 0));
        //    map2.Add(21, new Point(2, 1));
        //    map2.Add(22, new Point(3, 0));
        //    map2.Add(23, new Point(3, 1));
        //    map2.Add(24, new Point(4, 0));
        //    map2.Add(25, new Point(4, 1));
        //    map2.Add(26, new Point(5, 0));
        //    map2.Add(27, new Point(5, 1));
        //    map2.Add(28, new Point(6, 0));
        //    map2.Add(29, new Point(6, 1));
        //    map2.Add(30, new Point(7, 0));
        //    map2.Add(31, new Point(7, 1));
        //    map2.Add(32, new Point(8, 2));
        //    map2.Add(33, new Point(8, 3));
        //    map2.Add(34, new Point(9, 2));
        //    map2.Add(35, new Point(9, 3));
        //    map2.Add(36, new Point(10, 2));
        //    map2.Add(37, new Point(10, 3));
        //    map2.Add(38, new Point(11, 2));
        //    map2.Add(39, new Point(11, 3));
        //    map2.Add(40, new Point(12, 2));
        //    map2.Add(41, new Point(12, 3));
        //    map2.Add(42, new Point(13, 2));
        //    map2.Add(43, new Point(13, 3));
        //    map2.Add(44, new Point(14, 2));
        //    map2.Add(45, new Point(14, 3));
        //    map2.Add(46, new Point(15, 2));
        //    map2.Add(47, new Point(15, 3));
        //    map2.Add(48, new Point(8, 0));
        //    map2.Add(49, new Point(8, 1));
        //    map2.Add(50, new Point(9, 0));
        //    map2.Add(51, new Point(9, 1));
        //    map2.Add(52, new Point(10, 0));
        //    map2.Add(53, new Point(10, 1));
        //    map2.Add(54, new Point(11, 0));
        //    map2.Add(55, new Point(11, 1));
        //    map2.Add(56, new Point(12, 0));
        //    map2.Add(57, new Point(12, 1));
        //    map2.Add(58, new Point(13, 0));
        //    map2.Add(59, new Point(13, 1));
        //    map2.Add(60, new Point(14, 0));
        //    map2.Add(61, new Point(14, 1));
        //    map2.Add(62, new Point(15, 0));
        //    map2.Add(63, new Point(15, 1));

        //    int[] map = new int[]
        //    {                                
        //        0,  1,  8,  37,  2,  5,  6,  7,  4,  9,  10, 11, 12, 13, 14, 15,
        //                //
        //        16, 17, 52, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
        //        32, 33, 34, 35, 36, 3, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
        //                        //
        //        48, 49, 50, 51, 18, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63
        //    };

        //    int inPos = 0;
        //    int pageCount = imgWidth / 128;

        //    //for (int pageNum = 0; pageNum < pageCount; pageNum++)
        //    //{
        //    //    int pageSpace = pageNum * 128;
        //        for (int y = 0; y < imgHeight; y += 8)
        //        {
        //            for (int x = 0; x < imgWidth; x += 16)
        //            {
        //                for (int i = 0; i < 32; i++)
        //                {
        //                    int yPos = y;
        //                    int xPos = x;

        //                    Point mapPoint = (Point)map1[map[i]];

        //                    xPos += mapPoint.X;
        //                    yPos += mapPoint.Y;

        //                    outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                    //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                    inPos++;
        //                }
        //            }

        //            for (int x = 0; x < imgWidth; x += 16)
        //            {
        //                for (int i = 0; i < 32; i++)
        //                {
        //                    int yPos = y + 4;
        //                    int xPos = x;

        //                    Point mapPoint = (Point)map2[map[i]];

        //                    xPos += mapPoint.X;
        //                    yPos += mapPoint.Y;

        //                    outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                    //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                    inPos++;
        //                }
        //            }
        //            for (int x = 0; x < imgWidth; x += 16)
        //            {
        //                for (int i = 32; i < 64; i++)
        //                {
        //                    int yPos = y;
        //                    int xPos = x;

        //                    Point mapPoint = (Point)map1[map[i]];

        //                    xPos += mapPoint.X;
        //                    yPos += mapPoint.Y;

        //                    outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                    //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                    inPos++;
        //                }
        //            }

        //            for (int x = 0; x < imgWidth; x += 16)
        //            {
        //                for (int i = 32; i < 64; i++)
        //                {
        //                    int yPos = y + 4;
        //                    int xPos = x;

        //                    Point mapPoint = (Point)map2[map[i]];

        //                    xPos += mapPoint.X;
        //                    yPos += mapPoint.Y;

        //                    outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                    //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                    inPos++;
        //                }
        //            }
        //        }
        //    //}

        //    return outData;
        //}

        //protected byte[] unPS2DefianceSwizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        //{
        //    //return inData;
        //    byte[] outData;
        //    outData = new byte[inData.Length];

        //    //int[] masterMapX1 = new int[]
        //    //{
        //    //    0,  0,  1,  1,  2,  2,  3,  3,     4,   4,  5,  5,  6,  6,  7,  7,     
                
        //    //    8,   8,  9,  9, 10, 10, 11, 11,     12, 12, 13, 13, 14, 14, 15, 15
        //    //};

        //    //int[] masterMapY1 = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    //};


        //    //int[] masterMapX2 = new int[]
        //    //{
        //    //    0,  0,  1,  1,  2,  2,  3,  3,     4,   4,  5,  5,  6,  6,  7,  7,     
                
        //    //    8,   8,  9,  9, 10, 10, 11, 11,     12, 12, 13, 13, 14, 14, 15, 15
        //    //};

        //    //int[] masterMapY2 = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    //};

        //    //Hashtable map1 = new Hashtable();
        //    //Hashtable map2 = new Hashtable();

        //    //map1.Add(0, new Point(0, 0));
        //    //map1.Add(1, new Point(0, 1));
        //    //map1.Add(2, new Point(1, 0));
        //    //map1.Add(3, new Point(1, 1));
        //    //map1.Add(4, new Point(2, 0));
        //    //map1.Add(5, new Point(2, 1));
        //    //map1.Add(6, new Point(3, 0));
        //    //map1.Add(7, new Point(3, 1));
        //    //map1.Add(8, new Point(4, 0));
        //    //map1.Add(9, new Point(4, 1));
        //    //map1.Add(10, new Point(5, 0));
        //    //map1.Add(11, new Point(5, 1));
        //    //map1.Add(12, new Point(6, 0));
        //    //map1.Add(13, new Point(6, 1));
        //    //map1.Add(14, new Point(7, 0));
        //    //map1.Add(15, new Point(7, 1));
        //    //map1.Add(16, new Point(8, 0));
        //    //map1.Add(17, new Point(8, 1));
        //    //map1.Add(18, new Point(9, 0));
        //    //map1.Add(19, new Point(9, 1));
        //    //map1.Add(20, new Point(10, 0));
        //    //map1.Add(21, new Point(10, 1));
        //    //map1.Add(22, new Point(11, 0));
        //    //map1.Add(23, new Point(11, 1));
        //    //map1.Add(24, new Point(12, 0));
        //    //map1.Add(25, new Point(12, 1));
        //    //map1.Add(26, new Point(13, 0));
        //    //map1.Add(27, new Point(13, 1));
        //    //map1.Add(28, new Point(14, 0));
        //    //map1.Add(29, new Point(14, 1));
        //    //map1.Add(30, new Point(15, 0));
        //    //map1.Add(31, new Point(15, 1));

        //    //map2.Add(0, new Point(0, 0));
        //    //map2.Add(1, new Point(0, 1));
        //    //map2.Add(2, new Point(1, 0));
        //    //map2.Add(3, new Point(1, 1));
        //    //map2.Add(4, new Point(2, 0));
        //    //map2.Add(5, new Point(2, 1));
        //    //map2.Add(6, new Point(3, 0));
        //    //map2.Add(7, new Point(3, 1));
        //    //map2.Add(8, new Point(4, 0));
        //    //map2.Add(9, new Point(4, 1));
        //    //map2.Add(10, new Point(5, 0));
        //    //map2.Add(11, new Point(5, 1));
        //    //map2.Add(12, new Point(6, 0));
        //    //map2.Add(13, new Point(6, 1));
        //    //map2.Add(14, new Point(7, 0));
        //    //map2.Add(15, new Point(7, 1));
        //    //map2.Add(16, new Point(8, 0));
        //    //map2.Add(17, new Point(8, 1));
        //    //map2.Add(18, new Point(9, 0));
        //    //map2.Add(19, new Point(9, 1));
        //    //map2.Add(20, new Point(10, 0));
        //    //map2.Add(21, new Point(10, 1));
        //    //map2.Add(22, new Point(11, 0));
        //    //map2.Add(23, new Point(11, 1));
        //    //map2.Add(24, new Point(12, 0));
        //    //map2.Add(25, new Point(12, 1));
        //    //map2.Add(26, new Point(13, 0));
        //    //map2.Add(27, new Point(13, 1));
        //    //map2.Add(28, new Point(14, 0));
        //    //map2.Add(29, new Point(14, 1));
        //    //map2.Add(30, new Point(15, 0));
        //    //map2.Add(31, new Point(15, 1));

        //    Hashtable map1 = new Hashtable();
        //    Hashtable map2 = new Hashtable();

        //    //map1.Add(0, new Point(0, 0));
        //    //map1.Add(1, new Point(5, 1));
        //    //map1.Add(2, new Point(1, 0));
        //    //map1.Add(3, new Point(1, 1));
        //    //map1.Add(4, new Point(2, 0));
        //    //map1.Add(5, new Point(2, 1));
        //    //map1.Add(6, new Point(3, 0));
        //    //map1.Add(7, new Point(3, 1));
        //    //map1.Add(8, new Point(4, 0));
        //    //map1.Add(9, new Point(4, 1));
        //    //map1.Add(10, new Point(5, 0));
        //    //map1.Add(11, new Point(0, 1));
        //    //map1.Add(12, new Point(6, 0));
        //    //map1.Add(13, new Point(6, 1));
        //    //map1.Add(14, new Point(7, 0));
        //    //map1.Add(15, new Point(7, 1));
        //    //map1.Add(16, new Point(8, 0));
        //    //map1.Add(17, new Point(14, 1));
        //    //map1.Add(18, new Point(9, 0));
        //    //map1.Add(19, new Point(9, 1));
        //    //map1.Add(20, new Point(10, 0));
        //    //map1.Add(21, new Point(10, 1));
        //    //map1.Add(22, new Point(11, 0));
        //    //map1.Add(23, new Point(11, 1));
        //    //map1.Add(24, new Point(12, 0));
        //    //map1.Add(25, new Point(12, 1));
        //    //map1.Add(26, new Point(14, 0));
        //    //map1.Add(27, new Point(13, 1));
        //    //map1.Add(28, new Point(13, 0));
        //    //map1.Add(29, new Point(8, 1));
        //    //map1.Add(30, new Point(15, 0));
        //    //map1.Add(31, new Point(15, 1));

        //    //map1.Add(0, new Point(0, 0));
        //    //map1.Add(1, new Point(1, 0));
        //    //map1.Add(2, new Point(0, 1));
        //    //map1.Add(3, new Point(1, 1));
        //    //map1.Add(4, new Point(2, 0));
        //    //map1.Add(5, new Point(3, 0));
        //    //map1.Add(6, new Point(2, 1));
        //    //map1.Add(7, new Point(3, 1));
        //    //map1.Add(8, new Point(4, 0));
        //    //map1.Add(9, new Point(5, 0));
        //    //map1.Add(10, new Point(4, 1));
        //    //map1.Add(11, new Point(5, 1));
        //    //map1.Add(12, new Point(6, 0));
        //    //map1.Add(13, new Point(7, 0));
        //    //map1.Add(14, new Point(6, 1));
        //    //map1.Add(15, new Point(7, 1));
        //    //map1.Add(16, new Point(8, 0));
        //    //map1.Add(17, new Point(9, 0));
        //    //map1.Add(18, new Point(8, 1));
        //    //map1.Add(19, new Point(9, 1));
        //    //map1.Add(20, new Point(10, 0));
        //    //map1.Add(21, new Point(11, 0));
        //    //map1.Add(22, new Point(10, 1));
        //    //map1.Add(23, new Point(11, 1));
        //    //map1.Add(24, new Point(12, 0));
        //    //map1.Add(25, new Point(13, 0));
        //    //map1.Add(26, new Point(12, 1));
        //    //map1.Add(27, new Point(13, 1));
        //    //map1.Add(28, new Point(14, 0));
        //    //map1.Add(29, new Point(15, 0));
        //    //map1.Add(30, new Point(14, 1));
        //    //map1.Add(31, new Point(15, 1));

        //    map1.Add(0, new Point(0, 0));
        //    map1.Add(1, new Point(5, 1));
        //    map1.Add(2, new Point(4, 0));
        //    map1.Add(3, new Point(10, 1));
        //    map1.Add(4, new Point(1, 0));
        //    map1.Add(5, new Point(12, 1));//2
        //    map1.Add(6, new Point(3, 0));
        //    map1.Add(7, new Point(3, 1));
        //    map1.Add(8, new Point(2, 0));
        //    map1.Add(9, new Point(4, 1));
        //    map1.Add(10, new Point(5, 0));
        //    map1.Add(11, new Point(14, 1));//0
        //    map1.Add(12, new Point(6, 0));
        //    map1.Add(13, new Point(6, 1));
        //    map1.Add(14, new Point(7, 0));
        //    map1.Add(15, new Point(7, 1));
        //    map1.Add(16, new Point(8, 0));
        //    map1.Add(17, new Point(0, 1));//14
        //    map1.Add(18, new Point(9, 0));
        //    map1.Add(19, new Point(9, 1));
        //    map1.Add(20, new Point(10, 0));
        //    map1.Add(21, new Point(1, 1));
        //    map1.Add(22, new Point(11, 0));
        //    map1.Add(23, new Point(11, 1));
        //    map1.Add(24, new Point(12, 0));
        //    map1.Add(25, new Point(2, 1));//12
        //    map1.Add(26, new Point(14, 0));
        //    map1.Add(27, new Point(13, 1));
        //    map1.Add(28, new Point(13, 0));
        //    map1.Add(29, new Point(8, 1));
        //    map1.Add(30, new Point(15, 0));
        //    map1.Add(31, new Point(15, 1));

        //    map2.Add(0, new Point(5, 0));
        //    map2.Add(1, new Point(0, 1));
        //    map2.Add(2, new Point(10, 0));
        //    map2.Add(3, new Point(2, 1));
        //    map2.Add(4, new Point(2, 0));
        //    map2.Add(5, new Point(1, 1));
        //    map2.Add(6, new Point(3, 0));
        //    map2.Add(7, new Point(3, 1));
        //    map2.Add(8, new Point(4, 0));
        //    map2.Add(9, new Point(4, 1));
        //    map2.Add(10, new Point(0, 0));
        //    map2.Add(11, new Point(5, 1));
        //    map2.Add(12, new Point(6, 0));
        //    map2.Add(13, new Point(6, 1));
        //    map2.Add(14, new Point(7, 0));
        //    map2.Add(15, new Point(7, 1));
        //    map2.Add(16, new Point(14, 0));
        //    map2.Add(17, new Point(8, 1));
        //    map2.Add(18, new Point(9, 0));
        //    map2.Add(19, new Point(9, 1));
        //    map2.Add(20, new Point(1, 0));
        //    map2.Add(21, new Point(10, 1));
        //    map2.Add(22, new Point(11, 0));
        //    map2.Add(23, new Point(11, 1));
        //    map2.Add(24, new Point(12, 0));
        //    map2.Add(25, new Point(12, 1));
        //    map2.Add(26, new Point(13, 0));
        //    map2.Add(27, new Point(14, 1));
        //    map2.Add(28, new Point(8, 0));
        //    map2.Add(29, new Point(13, 1));
        //    map2.Add(30, new Point(15, 0));
        //    map2.Add(31, new Point(15, 1));

        //    int inPos = 0;

        //    for (int y = 0; y < imgHeight; y += 8)
        //    {
        //        for (int x = 0; x < (imgWidth * 2); x += 32)
        //        {
        //            for (int i = 0; i < 32; i++)
        //            {
        //                int yPos = y;
        //                int xPos = (x / 2);

        //                Point mapPoint = (Point)map1[i];

        //                xPos += mapPoint.X;
        //                yPos += mapPoint.Y;

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 32)
        //        {
        //            for (int i = 0; i < 32; i++)
        //            {
        //                int yPos = y + 2;
        //                int xPos = (x / 2);

        //                Point mapPoint = (Point)map1[i];

        //                xPos += mapPoint.X;
        //                yPos += mapPoint.Y;

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 32)
        //        {
        //            for (int i = 0; i < 32; i++)
        //            {
        //                int yPos = y + 4;
        //                int xPos = (x / 2);

        //                Point mapPoint = (Point)map2[i];

        //                xPos += mapPoint.X;
        //                yPos += mapPoint.Y;

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 32)
        //        {
        //            for (int i = 0; i < 32; i++)
        //            {
        //                int yPos = y + 6;
        //                int xPos = (x / 2);

        //                Point mapPoint = (Point)map2[i];

        //                xPos += mapPoint.X;
        //                yPos += mapPoint.Y;

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                //outData[inPos] = inData[(yPos * imgWidth) + xPos];
        //                inPos++;
        //            }
        //        }
        //    }

        //    return outData;
        //}

        //protected byte[] unPS2DefianceSwizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        //{
        //    //return inData;
        //    byte[] outData;
        //    outData = new byte[inData.Length];

        //    //int[] pattern1x = new int[]
        //    //{
        //    //    0,  1,  2,  3,  4,  5,  6,  7,
        //    //    0,  1,  2,  3,  4,  5,  6,  7
        //    //};

        //    //int[] pattern1y = new int[]
        //    //{
        //    //    0,  0,  0,  0,  0,  0,  0,  0,  
        //    //    1,  1,  1,  1,  1,  1,  1,  1
        //    //};

        //    //int[] pattern2x = new int[]
        //    //{
        //    //    1,  7,  2,  3,  4,  5,  6,  0,
        //    //    0,  3,  7,  1,  4,  5,  6,  2
        //    //};

        //    //int[] pattern2y = new int[]
        //    //{
        //    //    0,  0,  0,  0,  0,  0,  0,  0,  
        //    //    1,  1,  1,  1,  1,  1,  1,  1
        //    //};

        //    //int[] pattern1x = new int[]
        //    //{
        //    //    1,  7,  2,  3,  4,  5,  6,  0,
        //    //    0,  3,  7,  1,  4,  5,  6,  2
        //    //};

        //    //int[] pattern1y = new int[]
        //    //{
        //    //    0,  0,  0,  0,  0,  0,  0,  0,  
        //    //    1,  1,  1,  1,  1,  1,  1,  1
        //    //};

        //    //int[] pattern2x = new int[]
        //    //{
        //    //    6,  1,  4,  3,  7,  5,  2,  0,
        //    //    7,  3,  1,  0,  4,  5,  6,  2
        //    //};

        //    //int[] pattern2y = new int[]
        //    //{
        //    //    0,  0,  0,  0,  0,  0,  0,  0,  
        //    //    1,  1,  1,  1,  1,  1,  1,  1
        //    //};

        //    //int[] pattern1 = new int[]
        //    //{
        //    //    1,  5,  2,  4,  3,  6,  0,  7,  
        //    //    11,  14,  12,  8,  10,  13,  9,  15
        //    //};

        //    //int[] pattern2 = new int[]
        //    //{
        //    //    0,  1,  2,  3,  4,  5,  6,  7,  
        //    //    8,  9,  10,  11,  12,  13,  14,  15
        //    //};

        //    //int[] pattern1 = new int[]
        //    //{
        //    //    1,  0,
        //    //    3,  2,
        //    //    5,  4,
        //    //    7,  6,
        //    //    9,  8,
        //    //    11, 10,
        //    //    13, 12,
        //    //    15, 14
        //    //};

        //    //int[] pattern1 = new int[]
        //    //{
        //    //    0,  1,
        //    //    2,  3,
        //    //    4,  5,
        //    //    6,  7,
        //    //    8,  9,
        //    //    10, 11,
        //    //    12, 13,
        //    //    14, 15
        //    //};

        //    //int[] masterMapX = new int[]
        //    //{
        //    //    0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7
        //    //};

        //    //int[] masterMapY = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1
        //    //};

        //    //int[] pattern1 = new int[]
        //    //{
        //    //    0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15
        //    //};

        //    //int[] pattern2 = new int[]
        //    //{
        //    //    0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15
        //    //};


        //    //int[] masterMapX1 = new int[]
        //    //{
        //    //    0, 6, 5, 1, 2, 2, 3, 3,     4, 4, 6, 0, 1, 5, 7, 7
        //    //};

        //    //int[] masterMapY1 = new int[]
        //    //{
        //    //    0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    //};

        //    int[] masterMapX1 = new int[]
        //    {
        //        0, 6, 5, 1, 2, 2, 3, 3,     4, 4, 6, 0, 1, 5, 7, 7
        //    };

        //    int[] masterMapY1 = new int[]
        //    {
        //        0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    };


        //    int[] masterMapX2 = new int[]
        //    {
        //        6, 0, 1, 1, 2, 2, 3, 3,     4, 4, 0, 6, 5, 5, 7, 7
        //    };

        //    int[] masterMapY2 = new int[]
        //    {
        //        0, 1, 0, 1, 0, 1, 0, 1,     0, 1, 0, 1, 0, 1, 0, 1
        //    };

        //    int inPos = 0;

        //    for (int y = 0; y < imgHeight; y += 8)
        //    {
        //        for (int x = 0; x < (imgWidth * 2); x += 16)
        //        {
        //            for (int i = 0; i < 16; i++)
        //            {
        //                int yPos = y;
        //                int xPos = (x / 2);

        //                xPos += masterMapX1[i];
        //                yPos += masterMapY1[i];

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 16)
        //        {
        //            for (int i = 0; i < 16; i++)
        //            {
        //                int yPos = y + 2;
        //                int xPos = (x / 2);

        //                xPos += masterMapX1[i];
        //                yPos += masterMapY1[i];

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 16)
        //        {
        //            for (int i = 0; i < 16; i++)
        //            {
        //                int yPos = y + 4;
        //                int xPos = (x / 2);

        //                xPos += masterMapX2[i];
        //                yPos += masterMapY2[i];

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                inPos++;
        //            }
        //        }

        //        for (int x = 0; x < (imgWidth * 2); x += 16)
        //        {
        //            for (int i = 0; i < 16; i++)
        //            {
        //                int yPos = y + 6;
        //                int xPos = (x / 2);

        //                xPos += masterMapX2[i];
        //                yPos += masterMapY2[i];

        //                outData[(yPos * imgWidth) + xPos] = inData[inPos];
        //                inPos++;
        //            }
        //        }
        //    }

        //    return outData;
        //}

        protected int[] BuildPSMT8BlockIndex()
        {
            int[] blockIndex = new int[] 
            {
                0,  1,  4,  5,  16, 17, 20, 21,
                2,  3,  6,  7,  18, 19, 22, 23,
                8,  9,  12, 13, 24, 25, 28, 29,
                10, 11, 14, 15, 26, 27, 30, 31
            };
            return blockIndex;
        }

        protected int[] BuildPSMT8ColumnIndex()
        {
            int[] columnIndex = new int[]
            {
                0,
                1,
                2,
                3
            };
            return columnIndex;
        }

        protected int[] BuildPSMT8ByteIndex()
        {
            int[] byteIndex = new int[]
            {
                0,  0,  0,  0,  0,  0,  0,  0,      2,  2,  2,  2,  2,  2,  2,  2,
                0,  0,  0,  0,  0,  0,  0,  0,      2,  2,  2,  2,  2,  2,  2,  2,

                1,  1,  1,  1,  1,  1,  1,  1,      3,  3,  3,  3,  3,  3,  3,  3,
                1,  1,  1,  1,  1,  1,  1,  1,      3,  3,  3,  3,  3,  3,  3,  3
            };

            return byteIndex;
        }

        protected int[][] BuildPSMT8PixelIndex()
        {
            int[] pixelIndex02 = new int[]
            {
                0,  1,  4,  5,  8,  9,  12, 13,     0,  1,  4,  5,  8,  9,  12, 13,
                2,  3,  6,  7,  10, 11, 14, 15,     2,  3,  6,  7,  10, 11, 14, 15,

                8,  9,  12, 13, 0,  1,  4,  5,      8,  9,  12, 13, 0,  1,  4,  5, 
                10, 11, 14, 15, 2,  3,  6,  7,      10, 11, 14, 15, 2,  3,  6,  7
            };

            int[] pixelIndex13 = new int[]
            {
                8,  9,  12, 13, 0,  1,  4,  5,      8,  9,  12, 13, 0,  1,  4,  5, 
                10, 11, 14, 15, 2,  3,  6,  7,      10, 11, 14, 15, 2,  3,  6,  7,

                0,  1,  4,  5,  8,  9,  12, 13,     0,  1,  4,  5,  8,  9,  12, 13,
                2,  3,  6,  7,  10, 11, 14, 15,     2,  3,  6,  7,  10, 11, 14, 15
            };

            int[][] pixelIndex = new int[4][];
            pixelIndex[0] = pixelIndex02;
            pixelIndex[1] = pixelIndex13;
            pixelIndex[2] = pixelIndex02;
            pixelIndex[3] = pixelIndex13;

            //int[] byteIndex = BuildPSMT8ByteIndex();

            //int[] tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex02[j] + (16 * byteIndex[j]);
            //}

            //pixelIndex[0] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex13[j] + (16 * byteIndex[j]) + 64;
            //}

            //pixelIndex[1] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex02[j] + (16 * byteIndex[j]) + 128;
            //}

            //pixelIndex[2] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex13[j] + (16 * byteIndex[j]) + 192;
            //}

            //pixelIndex[3] = tempIndex;

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i] = pixelIndex02[i] + (16 * byteIndex[i]);
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 64] = pixelIndex13[i] + (16 * byteIndex[i]) + 64;
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 128] = pixelIndex02[i] + (16 * byteIndex[i]) + 128;
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 192] = pixelIndex13[i] + (16 * byteIndex[i]) + 192;
            ////}

            return pixelIndex;
        }

        protected int[] BuildIDTex8ClutIndex()
        {
            int[] clutIndex = new int[256];
            int currentIndex = 0;
            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 16 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 8 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 24 + j] = currentIndex;
                    currentIndex++;
                }
            }
            return clutIndex;
        }


        public string GetImageTypeName()
        {
            switch (mFileType)
            {
                case FileTypes.SRTex:
                    return FILEDESC_SRTEX;
                    break;
                case FileTypes.SR2_PS2_AirForge_ARGB:
                    return FILEDESC_SR2_PS2_AIRFORGE_ARGB;
                    break;
                case FileTypes.SR2_PS2_AirForge_Indexed:
                    return FILEDESC_SR2_PS2_AIRFORGE_INDEXED;
                    break;
                case FileTypes.SR2_PS2_ARGB:
                    return FILEDESC_SR2_PS2_ARGB;
                    break;
                case FileTypes.SR2_PS2_Indexed:
                    return FILEDESC_SR2_PS2_INDEXED;
                    break;
                case FileTypes.SR2_PC:
                    return FILEDESC_SR2_PC;
                    break;
                case FileTypes.Defiance_PS2_ARGB:
                    return FILEDESC_DEFIANCE_PS2_ARGB;
                    break;
                case FileTypes.Defiance_PS2_Indexed:
                    return FILEDESC_DEFIANCE_PS2_INDEXED;
                    break;
                case FileTypes.Defiance_PC:
                    return FILEDESC_DEFIANCE_PC;
                    break;
                case FileTypes.TRL_PS2:
                    return FILEDESC_TRL_PS2;
                    break;
                case FileTypes.TRL_PSP_ARGB:
                    return FILEDESC_TRL_PSP_ARGB;
                    break;
                case FileTypes.TRL_PSP_Indexed:
                    return FILEDESC_TRL_PSP_INDEXED;
                    break;
                default:
                    return FILEDESC_UNKNOWN;
                    break;
            }
        }

        #region File I/O

        protected byte[] ReadDataFromFile(string inFile, long offset, long numBytes)
        {
            FileStream fStream;
            BinaryReader bReader;

            byte[] readData = new byte[numBytes];

            for (int i = 0; i < numBytes; i++)
            {
                readData[i] = (byte)0;
            }

            if (File.Exists(inFile))
            {
                //try
                //{
                    fStream = new FileStream(inFile, FileMode.Open, FileAccess.Read);
                    bReader = new BinaryReader(fStream);
                    fStream.Seek(offset, SeekOrigin.Begin);
                    fStream.Read(readData, 0, (int)numBytes);

                    bReader.Close();
                    fStream.Close();
                    return readData;
                //}
                //catch (EndOfStreamException ex)
                //{
                //    //MessageBox.Show("Reached end of file " + inFile + " before reading enough bytes", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                //}
                //catch (IOException ex)
                //{
                //    //MessageBox.Show("Error reading file: " + inFile, "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                //}
            }
            else
            {
                //MessageBox.Show("File not found: " + inFile, "File Not Found", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            return readData;
        }

        protected byte[] ReadDefiancePCDataFromFile(string inFile, long offset, long numBytes)
        {
            //reads the proprietary RLE-compressed data
            FileStream fStream;
            BinaryReader bReader;
            uint byteBuffer = 0;

            byte[] readData = new byte[numBytes];

            if (File.Exists(inFile))
            {
                try
                {
                    fStream = new FileStream(inFile, FileMode.Open, FileAccess.Read);
                    bReader = new BinaryReader(fStream);
                    fStream.Seek(offset, SeekOrigin.Begin);
                    int i = 0;
                    do
                    {
                        byteBuffer = bReader.ReadUInt32();
                        int runLength = 0;
                        uint byte1 = byteBuffer;
                        uint byte2 = byteBuffer;
                        uint byte3 = byteBuffer;
                        uint byte4 = byteBuffer;

                        byte1 <<= 24;
                        byte1 >>= 24;
                        byte2 <<= 16;
                        byte2 >>= 24;
                        byte3 <<= 8;
                        byte3 >>= 24;
                        byte4 >>= 24;

                        uint runCount1 = byte1;
                        uint runCount2 = byte2;
                        uint runCount3 = byte3;
                        uint runCount4 = byte4;

                        byte1 &= 0xFE;
                        byte2 &= 0xFE;
                        byte3 &= 0xFE;
                        byte4 &= 0xFE;

                        runCount1 &= 1;
                        runCount2 &= 1;
                        runCount3 &= 1;
                        runCount4 &= 1;

                        runCount2 <<= 1;
                        runCount3 <<= 2;
                        runCount4 <<= 3;

                        runLength = (int)(runCount1 | runCount2 | runCount3 | runCount4);
                        runLength++;

                        for (int j = 0; j < runLength; j++)
                        {
                            int jFour = j * 4;
                            readData[i + jFour] = (byte)byte4;
                            readData[i + jFour + 1] = (byte)byte3;
                            readData[i + jFour + 2] = (byte)byte2;
                            readData[i + jFour + 3] = (byte)byte1;
                        }
                        i += (4 * runLength);
                    } while (i < numBytes);
                    //fStream.Read(readData, 0, (int)numBytes);

                    bReader.Close();
                    fStream.Close();
                    return readData;
                }
                catch (EndOfStreamException ex)
                {
                    //MessageBox.Show("Reached end of file " + inFile + " before reading enough bytes", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (IOException ex)
                {
                    //MessageBox.Show("Error reading file: " + inFile, "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            else
            {
                //MessageBox.Show("File not found: " + inFile, "File Not Found", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            return readData;
        }
        #endregion

        #region Utility Functions

        protected uint ReadUInt32FromFile(string filePath, long offset)
        {
            uint readUInt;
            FileStream iStream;
            BinaryReader iReader;

            iStream = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);
            iStream.Seek(offset, SeekOrigin.Begin);
            readUInt = iReader.ReadUInt32();
            iReader.Close();
            iStream.Close();
            return readUInt;
        }

        protected uint ReadMask(BinaryReader reader)
        {
            uint byte1 = reader.ReadByte();
            uint byte2 = reader.ReadByte();
            uint byte3 = reader.ReadByte();
            uint byte4 = reader.ReadByte();

            byte1 <<= 24;
            byte2 <<= 16;
            byte3 <<= 8;

            return byte1 | byte2 | byte3 | byte4;
        }

        //reverses a mask uint because the PC version of Defiance has backwards masks
        protected uint TransformMask(uint inMask)
        {
            uint byte1 = inMask & 0xFF;
            uint byte2 = inMask & 0xFF00;
            uint byte3 = inMask & 0xFF0000;
            uint byte4 = inMask & 0xFF000000;

            byte1 <<= 24;
            byte2 <<= 8;
            byte3 >>= 8;
            byte4 >>= 24;

            return byte1 | byte2 | byte3 | byte4;
        }

        protected int getBitPositionInDWord(uint inDWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 32; i++)
            {
                bitMask = (uint)(1 << (31 - i));
                masked = (inDWord & bitMask) >> (31 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

        protected int getBitPositionInWord(uint inWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 16; i++)
            {
                bitMask = (ushort)(1 << (15 - i));
                masked = (inWord & bitMask) >> (15 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

        #endregion

    }
}
