// The Eye of RAW
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Eye of RAW.

// The Eye of RAW is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Eye of RAW is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Eye of RAW (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32;

namespace EyeOfRAW
{
    public class ProgramSettings
    {
        //ignore alpha?
        protected bool mIgnoreAlpha;

        public bool IgnoreAlpha
        {
            get
            {
                return mIgnoreAlpha;
            }
            set
            {
                mIgnoreAlpha = value;
            }
        }

        public ProgramSettings()
        {
            ReadOptionsFromRegistry();
        }

        #region Registry I/O

        public void ReadOptionsFromRegistry()
        {
            RegistryKey regKey;
            regKey = Registry.CurrentUser;

            initRegistryKeys();

            regKey = regKey.OpenSubKey("Software\\The Lost Worlds\\Eye of RAW", true);
            if (regKey.GetValue("Ignore Alpha").ToString() == "0")
            {
                IgnoreAlpha = false;
            }
            else
            {
                IgnoreAlpha = true;
            }

        }

        public void WriteOptionsToRegistry()
        {
            RegistryKey regKey;
            regKey = Registry.CurrentUser;

            initRegistryKeys();

            regKey = regKey.OpenSubKey("Software\\The Lost Worlds\\Eye of RAW", true);

            //set the actual values
            if (IgnoreAlpha == false)
            {
                regKey.SetValue("Ignore Alpha", "0");
            }
            else
            {
                regKey.SetValue("Ignore Alpha", "1");
            }

        }

        protected void initRegistryKeys()
        {
            RegistryKey regKey;
            regKey = Registry.CurrentUser;
            bool foundIgnoreAlpha;

            regKey = regKey.OpenSubKey("Software\\The Lost Worlds\\Eye of RAW", true);
            if (regKey == null)
            {
                //create key and values
                regKey = Registry.CurrentUser;
                regKey = regKey.OpenSubKey("Software\\The Lost Worlds", true);
                if (regKey == null)
                {
                    //TLW key doesn't exist, create it
                    regKey = Registry.CurrentUser;
                    regKey = regKey.OpenSubKey("Software", true);
                    regKey.CreateSubKey("The Lost Worlds");
                    regKey = regKey.OpenSubKey("The Lost Worlds", true);
                }
                regKey.CreateSubKey("Eye of RAW");
                regKey = regKey.OpenSubKey("Eye of RAW", true);
            }

            //create any keys that are missing
            foundIgnoreAlpha = false;
            foreach (string keyName in regKey.GetValueNames())
            {
                if (keyName == "Ignore Alpha")
                {
                    foundIgnoreAlpha = true;
                }
            }

            if (foundIgnoreAlpha != true)
            {
                regKey.SetValue("Ignore Alpha", "0");
            }
        }
        #endregion
    }
}
